"""
vl53l5cx_py_wrapper_disable.pyx
Created by Perry Naseck on 2/1/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""


cdef class VL53L5CX:
    DEFAULT_ADDR = None
    DEFAULT_FREQ = None

    def __cinit__(self, dev_path: str,
                  target_addr = None,
                  freq = None) -> None:
        raise RuntimeError("vl53l5cx_py.driver not available on this platform!")



"""
__init__.py: Definitions for VL53L5CX drivers and helpers.

Created by Perry Naseck on 2/24/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""



"""
__init__.py: Definitions for sensors.

Created by Perry Naseck on 1/21/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

from arenarobot.service.service import ArenaRobotService


class ArenaRobotServiceSensor(ArenaRobotService):
    """Sensor class for the ARENA."""

    DEVICE_INSTANCE_TYPE = "sensor"
    DEVICE_INSTANCE_SENSOR_TYPE = "unknown"

    def __init__(self, device_instance_prefix="",
                 device_instance_sensor_type="unknown",
                 subtopic="",
                 **kwargs):
        """Initialize the sensor class."""
        self.device_instance_sensor_type = device_instance_sensor_type
        super().__init__(
            device_instance_type=ArenaRobotServiceSensor.DEVICE_INSTANCE_TYPE,
            device_instance_prefix=f"sensor_{device_instance_prefix}",
            subtopic=f"sensors/{subtopic}",
            **kwargs
        )

    def publish_msg_base(self):
        """Publish message base."""
        payload = super().publish_msg_base()
        payload.update({
            "device_instance_sensor_type": self.device_instance_sensor_type
        })
        return payload



"""
__init__.py: arenarobot scripts module root.

Created by Perry Naseck on 01/25/212

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""



#!/usr/bin/env python
# Alex Strasser Cylab 6/3/202

import serial
import io
import time
import numpy as np
from math import pi
import json
from datetime import date, datetime
from particlefilter import ParticleFilterLoc
#from arena import *
import pyrealsense2.pyrealsense2 as rs
#from scipy.spatial.transform import Rotation
UWB_BIAS = 0.2
UWB_STD = 0.1
IN_TO_METER = 0.0254
RAD_TO_DEG = 180 / pi
#scene = Scene(host="mqtt.arenaxr.org", scene="test0", realm="realm", namespace="astrasse")
#box = Box(object_id="my_box", position=Position(0, 0, 0), scale=Scale(0.3, 1.8, 0.3), color=(255, 0, 0))
#text = Text(text="User", position=Position(0, 2, 0), scale=Scale(3,3,3), parent=box)
pf = ParticleFilterLoc()
ser = serial.Serial('/dev/serial/by-id/usb-SEGGER_J-Link_000760025484-if00', 115200, timeout=1)
locations = {
  7: (166.75, 39.5, 270.75),
  8: (-7.5, 91.25, 104.5),
  9: (132.25, 57.675, 135.75),
  11: (41.25, 27.375, 0)
}
locations = {
  7: (1.828, 1.749, 2.558),
  8: (2.806, 0.799, -1.901),
  9: (-1.82, 0.168, 1.955),
  10: (-2.94, 2.804, -2.126),
  11: (1.334, 2.052, -3.678)
}
#for k in locations:
#  x, y, z = locations[k]
#  locations[k] = (x*IN_TO_METER, y*IN_TO_METER, z*IN_TO_METER)
print(locations)
#@scene.run_once
#def init():
#  scene.add_object(box)
#  scene.add_object(text)
# Declare RealSense pipeline, encapsulating the actual device and sensors
pipe = rs.pipeline()
# Build config object and request pose data
cfg = rs.config()
cfg.enable_stream(rs.stream.pose)
# Start streaming with requested config
pipe.start(cfg)
ser.readline()
try:
  data = None
  while True:
    line = ser.readline()
    if (line[0] != ord('#')):
      line = line[0:-2].replace(b' ', b'').decode('utf-8')
      #ID, RANGE, RSSI, TIMESTAMP
      curData = line.split(',')
      if (len(curData) < 4):
         print(curData)
         continue
      if int(curData[0]) in locations:
        x, y, z = locations[int(curData[0])]
        pf.depositRange(x, y, z, float(curData[1]), UWB_STD)
#      box.update_attributes(position=Position(x.value, y.value, z.value), rotation=Rotation(0, theta.value*RAD_TO_DEG, 0))
#      scene.update_object(box)
    frames = pipe.wait_for_frames()
    pose = frames.get_pose_frame()
    if pose:
      # Print some of the pose data to the terminal
      data = pose.get_pose_data()
#      print("Frame #{}".format(pose.frame_number))
#      print("Velocity: {}".format(data.velocity))
#      print("Acceleration: {}\n".format(data.acceleration))
    if data is not None:
      pf.depositVio(np.float64(time.time()), data.translation.x, data.translation.z * (-1), data.translation.y, 0.0)
      print("625_X:{: 3f} 625_Y:{: 3f} 625_Z:{: 3f}".format(data.translation.x,data.translation.y,data.translation.z))
    pf_data = pf.getTagLoc()
    print('PAR_X:{: 3f}'.format(pf_data[2]), 'PAR_Y:{: 3f}'.format(pf_data[3]), 'PAR_Z:{: 3f}'.format(pf_data[4]), 'PAR_T:{: 3f}'.format(pf_data[5]))
#    print(data)
finally:
  pipe.stop()


"""
vl53l5cx_py_wrapper.pyx
Created by Perry Naseck on 11/4/21.

Copyright (c) 2021, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""


import atexit
from libc.stdint cimport int32_t, uint8_t, uint16_t
import numpy as np
cimport numpy as np

from vl53l5cx_py.helpers import VL53L5CXSensorData

cdef extern from "../VL53L5CX_Linux_driver_1.1.2/user/uld-driver/inc/vl53l5cx_api.h":
    enum: VL53L5CX_DEFAULT_I2C_ADDRESS
    enum: VL53L5CX_RESOLUTION_8X8
    enum: VL53L5CX_NB_TARGET_PER_ZONE
    ctypedef struct VL53L5CX_Configuration:
        pass
    # ctypedef struct motion_indicator_t:
    #     uint32_t global_indicator_1
    #     uint32_t global_indicator_2
    #     uint8_t	 status
    #     uint8_t	 nb_of_detected_aggregates
    #     uint8_t	 nb_of_aggregates
    #     uint8_t	 spare
    #     uint32_t motion[32]
    ctypedef struct VL53L5CX_ResultsData:
        #ifndef VL53L5CX_DISABLE_AMBIENT_PER_SPAD
        np.uint32_t ambient_per_spad[VL53L5CX_RESOLUTION_8X8]
        #endif

        #ifndef VL53L5CX_DISABLE_NB_TARGET_DETECTED
        np.uint8_t nb_target_detected[VL53L5CX_RESOLUTION_8X8]
        #endif

        #ifndef VL53L5CX_DISABLE_NB_SPADS_ENABLED
        np.uint32_t nb_spads_enabled[VL53L5CX_RESOLUTION_8X8]
        #endif

        #ifndef VL53L5CX_DISABLE_SIGNAL_PER_SPAD
        np.uint32_t signal_per_spad[(VL53L5CX_RESOLUTION_8X8
                        *VL53L5CX_NB_TARGET_PER_ZONE)]
        #endif

        #ifndef VL53L5CX_DISABLE_RANGE_SIGMA_MM
        np.uint16_t range_sigma_mm[(VL53L5CX_RESOLUTION_8X8
                        *VL53L5CX_NB_TARGET_PER_ZONE)]
        #endif

        #ifndef VL53L5CX_DISABLE_DISTANCE_MM
        np.int16_t distance_mm[(VL53L5CX_RESOLUTION_8X8
                        *VL53L5CX_NB_TARGET_PER_ZONE)]
        #endif

        #ifndef VL53L5CX_DISABLE_REFLECTANCE_PERCENT
        np.uint8_t reflectance[(VL53L5CX_RESOLUTION_8X8
                        *VL53L5CX_NB_TARGET_PER_ZONE)]
        #endif

        #ifndef VL53L5CX_DISABLE_TARGET_STATUS
        np.uint8_t target_status[(VL53L5CX_RESOLUTION_8X8
                        *VL53L5CX_NB_TARGET_PER_ZONE)]
        #endif

        #ifndef VL53L5CX_DISABLE_MOTION_INDICATOR
        # motion_indicator_t motion_indicator
        #endif

cdef extern from "../include/vl53l5cx_py.h":
    int32_t vl53l5cx_py_init(VL53L5CX_Configuration* dev_conf,
                             const char *dev_path, uint16_t target_addr,
                             uint8_t freq)
    int32_t vl53l5cx_py_close(VL53L5CX_Configuration* dev_conf)
    int32_t vl53l5cx_py_start_ranging(VL53L5CX_Configuration* dev_conf)
    int32_t vl53l5cx_py_stop_ranging(VL53L5CX_Configuration* dev_conf)
    int32_t vl53l5cx_py_get_range(VL53L5CX_Configuration* dev_conf,
                                  VL53L5CX_ResultsData* results)

cdef VL53L5CX_PY_DEFAULT_ADDR = VL53L5CX_DEFAULT_I2C_ADDRESS
cdef VL53L5CX_PY_DEFAULT_FREQ = 15


cdef class VL53L5CX:
    cdef VL53L5CX_Configuration dev_conf
    cdef bint ranging

    DEFAULT_ADDR = np.uint16(VL53L5CX_PY_DEFAULT_ADDR)
    DEFAULT_FREQ = np.uint8(VL53L5CX_PY_DEFAULT_FREQ)

    def __cinit__(self, dev_path: str,
                  target_addr: np.uint16_t = VL53L5CX_PY_DEFAULT_ADDR,
                  freq: np.uint8_t = VL53L5CX_PY_DEFAULT_FREQ) -> None:
        cdef int32_t status = vl53l5cx_py_init(&self.dev_conf, dev_path.encode(), target_addr, freq)
        if status != 0:
            raise RuntimeError
        self.ranging = False
        atexit.register(self.__del__)

    def __del__(self) -> None:
        if self.ranging:
            self.stop_ranging()
        cdef int32_t status = vl53l5cx_py_close(&self.dev_conf)
        if status != 0:
            raise RuntimeError
        atexit.unregister(self.__del__)

    cpdef void start_ranging(self):
        cdef int32_t status = vl53l5cx_py_start_ranging(&self.dev_conf)
        if status != 0:
            raise RuntimeError
        self.ranging = True

    cpdef void stop_ranging(self):
        cdef int32_t status = vl53l5cx_py_stop_ranging(&self.dev_conf)
        if status != 0:
            raise RuntimeError
        self.ranging = False

    cpdef get_range(self):
        cdef VL53L5CX_ResultsData results
        cdef int32_t status = vl53l5cx_py_get_range(&self.dev_conf, &results)
        if status != 0:
            raise RuntimeError

        return VL53L5CXSensorData(
            distance_mm=results.distance_mm,
            target_status=np.frombuffer(results.target_status, dtype='S1')
        )



"""
__init__.py: arenavideocall scripts module root.

Created by Perry Naseck on 11/3/21.

Copyright (c) 2021, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""



import matplotlib.pyplot as plt
import sys
from matplotlib.ticker import StrMethodFormatter, MaxNLocator

fig = plt.figure()

param = "Z"
if len(sys.argv) > 1:
  param = sys.argv[1]

pfPosFile = open("./data/pfOut.csv", "r")
arduAttFile = open("./data/arduPilotAtt.csv", "r")
arduGpsFile = open("./data/arduPilotGps.csv", "r")
t265File = open("./data/t265Pos.csv", "r")

index = "XYZRPW".index(param)

pfPos = [[], [], [], []]
arduAtt = [[], [], [], []]
arduGps = [[], [], [], []]
for line in pfPosFile.readlines():
  s = line.split(",")
  pfPos[0].append(float(s[0]))
  pfPos[1].append(float(s[1])-0.15)
  pfPos[2].append(-float(s[2]))
  pfPos[3].append(-float(s[3]))
for line in arduAttFile.readlines():
  s = line.split(",")
  arduAtt[0].append(float(s[0]))
  arduAtt[1].append(float(s[1]))
  arduAtt[2].append(float(s[2]))
  arduAtt[3].append(float(s[3]))
for line in arduGpsFile.readlines():
  s = line.split(",")
  arduGps[0].append(float(s[0]))
  arduGps[1].append(float(s[1]))
  arduGps[2].append(float(s[2]))
  arduGps[3].append(float(s[3]))

t265 = [[], [], [], [], [], [], []]
for line in t265File:
  s = line.split(",")
  t265[0].append(float(s[0]))
  t265[1].append(float(s[1]))
  t265[2].append(float(s[2]))
  t265[3].append(float(s[3]))
  t265[4].append(float(s[4]))
  t265[5].append(float(s[5]))
  t265[6].append(float(s[6]))

time_init = arduGps[0][0]
pfPos[0] = [i - time_init for i in pfPos[0]]
arduAtt[0] = [i - time_init for i in arduAtt[0]]
arduGps[0] = [i - time_init for i in arduGps[0]]
t265[0] = [i - time_init for i in t265[0]]

if len(sys.argv) > 1 and len(sys.argv[1]) > 1:
  ax = fig.add_subplot(projection='3d')
  ax.plot3D(t265[1], t265[2], t265[3])
  ax.plot3D(pfPos[1], pfPos[2], pfPos[3])
  ax.plot3D(arduGps[1], arduGps[2], arduGps[3])

  ax.legend()
  ax.set_xlabel('X Position')
  ax.set_ylabel('Y Position')
  ax.set_zlabel('Z Position')
  plt.show()

else:
  ax = fig.add_subplot()
  ax.yaxis.set_major_formatter(StrMethodFormatter('{x:,.2f}')) # 2 decimal places
  ax.yaxis.set_major_locator(MaxNLocator(5))
  ax.plot(t265[0], t265[index+1], label='T265')
  if index > 2:
    ax.plot(arduAtt[0], arduAtt[index-3+1], label='arduPilot')
  else:
    ax.plot(pfPos[0], pfPos[index+1], label='Particle Filter')
    ax.plot(arduGps[0], arduGps[index+1], label='arduPilot GPS')

  plt.xlabel("time (seconds)")
  plt.ylabel("distance (meters)")

  plt.xlabel("time (seconds)")
  plt.ylabel("distance (meters)")
  plt.title(param+" Position")

  ax.legend()
  plt.show()



#!/usr/bin/env python3

#####################################################
##          librealsense T265 to MAVLink           ##
#####################################################
# This script assumes pyrealsense2.[].so file is found under the same directory as this script
# Install required packages: 
#   pip3 install pyrealsense2
#   pip3 install transformations
#   pip3 install pymavlink
#   pip3 install apscheduler
#   pip3 install pyserial
#   pip3 install slam3d
#   pip3 install arena-py

# Set the path for IDLE
import sys
sys.path.append("/usr/local/lib/")

# Set MAVLink protocol to 2.
import os
os.environ["MAVLINK20"] = "1"

# Import the libraries
import pyrealsense2.pyrealsense2 as rs
import numpy as np
import transformations as tf
import math as m
import time
import argparse
import threading
import signal
import serial

from time import sleep
from apscheduler.schedulers.background import BackgroundScheduler
from pymavlink import mavutil

from particlefilter import ParticleFilterLoc
from arena import *

UWB_BIAS = 0.2
UWB_STD = 0.1

GPS_FIX_TYPE_PPP = 8
UINT16_MAX = 65535
UINT8_MAX = 255

IN_TO_METER = 0.0254
RAD_TO_DEG = 180 / m.pi
METER_TO_LATLNG = 10000000 / 111139

# Replacement of the standard print() function to flush the output
def progress(string):
    print(string, file=sys.stdout)
    sys.stdout.flush()

#######################################
# Parameters
#######################################

# Default configurations for connection to the FCU
connection_string_default = '/dev/ttyS0'
connection_baudrate_default = 115200

# Transformation to convert different camera orientations to NED convention. Replace camera_orientation_default for your configuration.
#   0: Forward, USB port to the right
#   1: Downfacing, USB port to the right 
#   2: Forward, 45 degree tilted down
# Important note for downfacing camera: you need to tilt the vehicle's nose up a little - not flat - before you run the script, otherwise the initial yaw will be randomized, read here for more details: https://github.com/IntelRealSense/librealsense/issues/4080. Tilt the vehicle to any other sides and the yaw might not be as stable.
camera_orientation_default = 3

# https://mavlink.io/en/messages/common.html#VISION_POSITION_ESTIMATE
enable_msg_vision_position_estimate = True
vision_position_estimate_msg_hz_default = 30.0

# https://mavlink.io/en/messages/common.html#VISION_SPEED_ESTIMATE
enable_msg_vision_speed_estimate = True
vision_speed_estimate_msg_hz_default = 30.0

# Monitor user's online input via keyboard, can only be used when runs from terminal
enable_user_keyboard_input = True

# Default global position for EKF home/ origin
enable_auto_set_ekf_home = True
home_lat = 0     # Somewhere random
home_lon = 0     # Somewhere random
home_alt = 0        # Somewhere random

# TODO: Taken care of by ArduPilot, so can be removed (once the handling on AP side is confirmed stable)
# In NED frame, offset from the IMU or the center of gravity to the camera's origin point
body_offset_enabled = 0
body_offset_x = 0  # In meters (m)
body_offset_y = 0  # In meters (m)
body_offset_z = 0  # In meters (m)

# Global scale factor, position x y z will be scaled up/down by this factor
scale_factor = 0.85

# Enable using yaw from compass to align north (zero degree is facing north)
compass_enabled = 1

# pose data confidence: 0x0 - Failed / 0x1 - Low / 0x2 - Medium / 0x3 - High 
pose_data_confidence_level = ('FAILED', 'Low', 'Medium', 'High')

# lock for thread synchronization
drone_lock = threading.Lock()
threads_should_exit = False

# default exit code is failure - a graceful termination with a
# terminate signal is possible.
exit_code = 1


#######################################
# Global variables
#######################################

# File variables
arduAttFile = open("./data/arduPilotAtt.csv", "w")
arduGlobalFile = open("./data/arduPilotGps.csv", "w")
cameraFile = open("./data/t265Pos.csv", "w")
imuFile = open("./data/imuData.csv", "w")
pfOutFile = open("./data/pfOut.csv", "w")
pfLocFile = open("./data/pfLoc.csv", "w")

# FCU connection variables

# Camera-related variables
pipe = None
pose_sensor = None
rs_cfg = None
linear_accel_cov = 0.0001
angular_vel_cov  = 0.0001

# Data variables
data = None
prev_data = None
H_aeroRef_aeroBody = None
V_aeroRef_aeroBody = None
heading_north_yaw = 0
current_global_pos = [0, 0, 0]
current_confidence_level = None
current_time_us = 0
home_msg_received = False
drone_ready = False
realsense_resetting = False
pfStatus = 0
alt_offset = 0

# Increment everytime pose_jumping or relocalization happens
# See here: https://github.com/IntelRealSense/librealsense/blob/master/doc/t265.md#are-there-any-t265-specific-options
# For AP, a non-zero "reset_counter" would mean that we could be sure that the user's setup was using mavlink2
reset_counter = 1

#######################################
# Parsing user' inputs
#######################################

parser = argparse.ArgumentParser(description='Reboots vehicle')
parser.add_argument('--connect', default=connection_string_default,
                    help="Vehicle connection target string. If not specified, a default string will be used.")
parser.add_argument('--baudrate', type=float, default=connection_baudrate_default,
                    help="Vehicle connection baudrate. If not specified, a default value will be used.")
parser.add_argument('--vision_position_estimate_msg_hz', type=float, default=vision_position_estimate_msg_hz_default,
                    help="Update frequency for VISION_POSITION_ESTIMATE message. If not specified, a default value will be used.")
parser.add_argument('--vision_speed_estimate_msg_hz', type=float, default=vision_speed_estimate_msg_hz_default,
                    help="Update frequency for VISION_SPEED_DELTA message. If not specified, a default value will be used.")
parser.add_argument('--scale_calib_enable', default=False, action='store_true',
                    help="Scale calibration. Only run while NOT in flight")
parser.add_argument('--camera_orientation', type=int, default=camera_orientation_default,
                    help="Configuration for camera orientation. Currently supported: forward, usb port to the right - 0; downward, usb port to the right - 1, 2: forward tilted down 45deg")
parser.add_argument('--arena_disable', action='store_true',
                    help="Disable streaming data to ARENA.")
parser.add_argument('--update_arena_hz', type=float, default=5,
                    help="Update frequency for position updates to ARENA. If not specified, a default value will be used.")
parser.add_argument('--uwb_disable', action='store_true',
                    help="Disable UWB")
parser.add_argument('--debug_enable', action='store_true',
                    help="Enable debug messages on terminal")

args = parser.parse_args()

connection_string = args.connect
connection_baudrate = args.baudrate
vision_position_estimate_msg_hz = args.vision_position_estimate_msg_hz
vision_speed_estimate_msg_hz = args.vision_speed_estimate_msg_hz
scale_calib_enable = args.scale_calib_enable
camera_orientation = args.camera_orientation
debug_enable = args.debug_enable
enable_arena = not args.arena_disable
update_arena_hz = args.update_arena_hz
enable_uwb = not args.uwb_disable


if body_offset_enabled == 1:
    progress("INFO: Using camera position offset: Enabled, x y z is %s %s %s" % (body_offset_x, body_offset_y, body_offset_z))
else:
    progress("INFO: Using camera position offset: Disabled")

if compass_enabled == 1:
    progress("INFO: Using compass: Enabled. Heading will be aligned to north.")
else:
    progress("INFO: Using compass: Disabled")

if scale_calib_enable == True:
    progress("\nINFO: SCALE CALIBRATION PROCESS. DO NOT RUN DURING FLIGHT.\nINFO: TYPE IN NEW SCALE IN FLOATING POINT FORMAT\n")
else:
    if scale_factor == 1.0:
        progress("INFO: Using default scale factor %s" % scale_factor)
    else:
        progress("INFO: Using scale factor %s" % scale_factor)

if camera_orientation == 0:     # Forward, USB port to the right
    H_aeroRef_T265Ref   = np.array([[0,0,-1,0],[1,0,0,0],[0,-1,0,0],[0,0,0,1]])
    H_T265body_aeroBody = np.linalg.inv(H_aeroRef_T265Ref)
elif camera_orientation == 1:   # Downfacing, USB port to the right
    H_aeroRef_T265Ref   = np.array([[0,0,-1,0],[1,0,0,0],[0,-1,0,0],[0,0,0,1]])
    H_T265body_aeroBody = np.array([[0,1,0,0],[1,0,0,0],[0,0,-1,0],[0,0,0,1]])
elif camera_orientation == 2:   # 45degree forward
    H_aeroRef_T265Ref   = np.array([[0,0,-1,0],[1,0,0,0],[0,-1,0,0],[0,0,0,1]])
    H_T265body_aeroBody = (tf.euler_matrix(m.pi/4, 0, 0)).dot(np.linalg.inv(H_aeroRef_T265Ref))
elif camera_orientation == 3:   # Upward facing, USB port to the right
    H_aeroRef_T265Ref   = np.array([[0,0,-1,0],[1,0,0,0],[0,-1,0,0],[0,0,0,1]])
    H_T265body_aeroBody = np.array([[0,1,0,0],[-1,0,0,0],[0,0,-1,0],[0,0,0,1]])
# elif camera_orientation == 4:   # Upward facing
#     H_aeroRef_T265Ref   = np.array([[0,0,1,0],[-1,0,0,0],[0,-1,0,0],[0,0,0,1]])
#     H_T265body_aeroBody = np.array([[0,1,0,0],[-1,0,0,0],[0,0,-1,0],[0,0,0,1]])
else:                           # Default is facing forward, USB port to the right
    H_aeroRef_T265Ref   = np.array([[0,0,-1,0],[1,0,0,0],[0,-1,0,0],[0,0,0,1]])
    H_T265body_aeroBody = np.linalg.inv(H_aeroRef_T265Ref)

if not debug_enable:
    debug_enable = 0
else:
    debug_enable = 1
    np.set_printoptions(precision=4, suppress=True) # Format output on terminal 
    progress("INFO: Debug messages enabled.")

#######################################
# ARENA, Particle Filter, & UWB
#######################################

arena_attitude = [0, 0, 0]
arena_batt = 0
if enable_arena:
    scene = Scene(host="mqtt.arenaxr.org", scene="ARENA-Drone", realm="realm", namespace="astrasse")
    arena_drone = Box(object_id="arena_drone", depth=0.4, width=0.4, height=0.07, position=Position(0, 0, 0), scale=Scale(1, 1, 1), color=(128, 0, 172), material={"opacity":0.5})
    scene.add_object(arena_drone)

def update_arena_pos():
    global arena_drone, arena_attitude, arena_batt, pfStatus
    with pf_lock:
        if pfStatus < 3:
            return
        pf_data = pf.getTagLoc()
    print("ARENA", pf_data[2], pf_data[4], -pf_data[3], pf_data[5])
    pfOutFile.write(f"{str(time.time())},{str(pf_data[1])},{str(pf_data[2])},{str(pf_data[3])},{str(pf_data[4])},{str(pf_data[5])}\n")
    arena_drone.update_attributes(position=Position(pf_data[2], pf_data[4], -pf_data[3]), rotation=Rotation(arena_attitude[0], arena_attitude[1]+pf_data[5]*RAD_TO_DEG, arena_attitude[2]))#, battery=Attribute(battery=arena_batt))
    scene.update_object(arena_drone)

uwb_locations = {}

def arena_uwb_location_update(tag):
    global uwb_locations
    pos = tag.data.position
    uwb_locations[int(tag.data.armarker.markerid)] = [pos.x, pos.y, pos.z]

def arena_get_uwb_locations():
    global uwb_locations
    print("GETTING UWB LOCATIONS FROM ARENA")
    for i in scene.all_objects:
        if ('armarker' in scene.all_objects[i].data and scene.all_objects[i].data.armarker.markertype == "uwb"):
            scene.all_objects[i].update_handler = arena_uwb_location_update
            arena_uwb_location_update(scene.all_objects[i])
    print(uwb_locations)

def arena_target_location_update(target):
    pos = target.data.position
    yaw = target.data.rotation.y
    print("***************************************************")
    print("TARGET UPDATE:", pos, yaw)
    if (pos.x == 0 and pos.y == 0 and pos.z == 0):
        land()
    else:
        set_target(pos.x, pos.z, -pos.y, yaw=yaw)

def arena_get_target_location():
    target = scene.all_objects["drone_target"]
    if target:
        target.update_handler = arena_target_location_update
    else:
        print("No target found, waypoint flying disabled")

if enable_uwb:
    pf_lock = threading.Lock()
    pf = ParticleFilterLoc()
    uwb = serial.Serial('/dev/ttyAMA1', 115200, timeout=1)

    uwb.write(b"AT+RATE 10\n")
    if not enable_arena:
        uwb_locations = {
            7: [1.988, 1.749, 2.608],
            8: [2.966, 0.799, -1.851],
            9: [-1.66, 0.168, 2.005],
            10: [-2.78, 2.804, -2.076],
            11: [1.494, 2.052, -3.628],
            13: [-0.324, 3.137, -0.44]
        }

####################################
# Functions - MAVLink
####################################

def uwb_loop():
    global uwb, uwb_locations, threads_should_exit, pfStatus
    #First line is usually unreadable
    uwb.readline()
    while not threads_should_exit:
        line = uwb.readline()
        if (len(line) > 5 and line[0] != ord('#')):
            line = line[0:-2].replace(b' ', b'').decode('utf-8')
            #ID, RANGE, RSSI, TIMESTAMP
            curData = line.split(',')
            if (len(curData) < 4):
                continue
            try:
                int(curData[0])
            except ValueError:
                continue
            if int(curData[0]) in uwb_locations.keys():
                x, y, z = uwb_locations[int(curData[0])]
                with pf_lock:
                    try:
                        # print("RANGE", np.float32(x), np.float32(-z), np.float32(y), np.float32(float(curData[1])))
                        pfLocFile.write(f"UWB,{time.time()},{int(curData[0])},{str(x)},{str(-z)},{str(y)},{str(float(curData[1]))},0.1\n")
                        pf.depositRange(np.float32(x), np.float32(-z), np.float32(y), np.float32(float(curData[1])), 0.1)
                        pfStatus |= 1
                    except:
                        pass

def mavlink_loop():
    '''a main routine for a thread; reads data from a mavlink connection,
    calling callbacks based on message type received.
    '''
    global conn, mavlink_callbacks, threads_should_exit
    print("Mavlink")
    interesting_messages = list(mavlink_callbacks.keys())
    while not threads_should_exit:
        m = conn.recv_match(type=interesting_messages, timeout=1, blocking=True)
        if m is None:
            continue
        # print(m.get_type())
        mavlink_callbacks[m.get_type()](m)

# https://mavlink.io/en/messages/common.html#VISION_POSITION_ESTIMATE
def send_vision_position_estimate_message():
    global current_time_us, H_aeroRef_aeroBody, reset_counter, current_confidence_level, realsense_resetting
    with drone_lock:
        if H_aeroRef_aeroBody is not None:
            cur_pos = [H_aeroRef_aeroBody[0][3], H_aeroRef_aeroBody[1][3], H_aeroRef_aeroBody[2][3]]

            # Setup angle data
            rpy_rad = np.array( tf.euler_from_matrix(H_aeroRef_aeroBody, 'sxyz'))

            rpy_rad[0] = m.pi - rpy_rad[0]
            if rpy_rad[0] > m.pi: rpy_rad[0] -= 2*m.pi

            # Setup covariance data, which is the upper right triangle of the covariance matrix, see here: https://files.gitter.im/ArduPilot/VisionProjects/1DpU/image.png
            # Attemp #01: following this formula https://github.com/IntelRealSense/realsense-ros/blob/development/realsense2_camera/src/base_realsense_node.cpp#L1406-L1411
            cov_pose    = linear_accel_cov * pow(10, 3 - int(3*current_confidence_level/100))
            cov_twist   = angular_vel_cov  * pow(10, 1 - int(3*current_confidence_level/100))

            if realsense_resetting:
                cov_pose *= 100
                cov_twist *= 100
            
            covariance  = np.array([cov_pose, 0, 0, 0, 0, 0,
                                       cov_pose, 0, 0, 0, 0,
                                          cov_pose, 0, 0, 0,
                                            cov_twist, 0, 0,
                                               cov_twist, 0,
                                                  cov_twist])

            # print(f"Sending: {cur_pos[0]}, {cur_pos[1]}, {cur_pos[2]}")
            # Send the message
            conn.mav.vision_position_estimate_send(
                current_time_us,            # us Timestamp (UNIX time or time since system boot)
                cur_pos[0],   # Global X position
                cur_pos[1],   # Global Y position
                cur_pos[2],   # Global Z position
                rpy_rad[0],                 # Roll angle
                rpy_rad[1],                 # Pitch angle
                rpy_rad[2],                 # Yaw angle
                covariance,                 # Row-major representation of pose 6x6 cross-covariance matrix
                reset_counter               # Estimate reset counter. Increment every time pose estimate jumps.
            )

            s = str(time.time()) + "," + str(cur_pos[0]) + "," + str(cur_pos[1]) + "," + str(cur_pos[2]) + "," + \
                        str(rpy_rad[0]) + "," + str(rpy_rad[1]) + "," + str(rpy_rad[2]) + "\n"
            cameraFile.write(s)

# https://mavlink.io/en/messages/common.html#VISION_SPEED_ESTIMATE
def send_vision_speed_estimate_message():
    global current_time_us, V_aeroRef_aeroBody, reset_counter, current_confidence_level
    with drone_lock:
        if V_aeroRef_aeroBody is not None:

            # Attemp #01: following this formula https://github.com/IntelRealSense/realsense-ros/blob/development/realsense2_camera/src/base_realsense_node.cpp#L1406-L1411
            cov_pose    = linear_accel_cov * pow(10, 3 - int(3*current_confidence_level/100))
            covariance  = np.array([cov_pose,   0,          0,
                                    0,          cov_pose,   0,
                                    0,          0,          cov_pose])
            
            # Send the message
            conn.mav.vision_speed_estimate_send(
                current_time_us,            # us Timestamp (UNIX time or time since system boot)
                V_aeroRef_aeroBody[0][3],   # Global X speed
                V_aeroRef_aeroBody[1][3],   # Global Y speed
                V_aeroRef_aeroBody[2][3],   # Global Z speed
                covariance,                 # covariance
                reset_counter               # Estimate reset counter. Increment every time pose estimate jumps.
            )


# Send a mavlink SET_GPS_GLOBAL_ORIGIN message (http://mavlink.org/messages/common#SET_GPS_GLOBAL_ORIGIN), which allows us to use local position information without a GPS.
def set_default_global_origin():
    with drone_lock:
        print("SENDING GLOBAL ORIGIN")
        conn.mav.set_gps_global_origin_send(
            1,
            home_lat, 
            home_lon,
            home_alt
        )

# Send a mavlink SET_HOME_POSITION message (http://mavlink.org/messages/common#SET_HOME_POSITION), which allows us to use local position information without a GPS.
def set_default_home_position():
    x = 0
    y = 0
    z = 0
    q = [1, 0, 0, 0]   # w x y z

    approach_x = 0
    approach_y = 0
    approach_z = 1

    with drone_lock:
        conn.mav.set_home_position_send(
            1,
            home_lat, 
            home_lon,
            home_alt,
            x,
            y,
            z,
            q,
            approach_x,
            approach_y,
            approach_z
        )

def set_mode(mode):
    mode_map = conn.mode_mapping()
    if mode in mode_map:
        mode = mode_map.get(mode)
    else:
        print("MODE NOT FOUND")
        return
    conn.mav.command_long_send(
        1,
        1,
        mavutil.mavlink.MAV_CMD_DO_SET_MODE,
        1,
        mavutil.mavlink.MAV_MODE_FLAG_CUSTOM_MODE_ENABLED,
        mode,
        0,
        0,
        0,
        0,
        0
    )


def set_target(north, east, down, yaw=0, timeout=None):
    alt = (-down)
    with drone_lock:
        conn.mav.mission_item_int_send(
            1,
            1,
            0,
            mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
            mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,
            2,
            1,
            0,
            0,
            0,
            yaw,
            int(north*METER_TO_LATLNG),
            int(east*METER_TO_LATLNG),
            alt
        )

    thresh = 0.1
    def condition():
        return abs(current_global_pos[0] - north) < thresh and \
                abs(current_global_pos[1] - east) < thresh and \
                abs(current_global_pos[2] + down) < thresh # down is plus since it's negative altitude
    wait_for(condition, timeout=timeout)
        
#speed_type:
#   0 is airspeed
#   1 is groundspeed
#   2 is climbspeed
#   3 is descentspeed
def set_drone_speed(speed_type, speed):
    conn.mav.command_long_send(
        1,
        1,
        mavutil.mavlink.MAV_CMD_DO_CHANGE_SPEED,
        0,
        speed_type,
        speed,
        -1,
        0,
        0,
        0,
        0
    )

def takeoff(alt):
    set_mode("GUIDED")
    conn.arducopter_arm()
    time.sleep(0.5)
    conn.mav.command_long_send(0, 0, mavutil.mavlink.MAV_CMD_NAV_TAKEOFF,
                                               0, 0, 0, 0, 0, 0, 0, alt)
    def check_alt():
        return current_global_pos[2] > alt - 0.2
    wait_for(check_alt)
    

def land():
    set_mode("LAND")
    def check_alt():
        return current_global_pos[2] < 0.3
    wait_for(check_alt)
    time.sleep(1)
    conn.arducopter_disarm()
    
def wait_for(condition, interval=0.1, timeout=None):
    startTime = time.time()
    while True:
        if main_loop_should_quit:
            return
        if condition():
            return
        if timeout != None and (time.time() - startTime > timeout):
            return
        time.sleep(0.1)

def reboot():
    conn.mav.command_long_send(
        1,
        1,
        mavutil.mavlink.MAV_CMD_PREFLIGHT_REBOOT_SHUTDOWN,
        0,
        1,
        0,
        0,
        0,
        0,
        0,
        0
    )

# Listen to attitude data to acquire heading when compass data is enabled
def att_msg_callback(value):
    global heading_north_yaw, arena_attitude
    arduAttFile.write(str(time.time()) + "," + str(value.roll) + "," + str(value.pitch) + "," + str(value.yaw) + "\n")
    # print(value)
    if heading_north_yaw is None:
        heading_north_yaw = value.yaw
        progress("INFO: Received first ATTITUDE message with heading yaw %.2f degrees" % m.degrees(heading_north_yaw))
    if enable_arena:
        arena_attitude[0] = value.roll*RAD_TO_DEG
        arena_attitude[1] = -value.yaw*RAD_TO_DEG
        arena_attitude[2] = value.pitch*RAD_TO_DEG

def global_msg_callback(value):
    global H_aeroRef_aeroBody, current_global_pos
    current_global_pos = [value.lat / METER_TO_LATLNG, value.lon / METER_TO_LATLNG, value.alt / 1000]
    arduGlobalFile.write(str(time.time()) + "," + str(value.lat / METER_TO_LATLNG) + "," + str(value.lon / METER_TO_LATLNG) + "," + str(-value.alt / 1000) + "\n")
    # with pf_lock:
    #     pf.depositVio(np.float64(time.time()), np.float32(value.lat / METER_TO_LATLNG), np.float32(-value.lon / METER_TO_LATLNG), np.float32(value.alt / 1000), 0)
    print(f"GLOBAL: {str(value.lat / METER_TO_LATLNG)}, {str(value.lon / METER_TO_LATLNG)}, {str(-value.alt / 1000)}")
    #vx, vy, vz, hdg

def local_msg_callback(value):
    print(value)

def home_msg_callback(value):
    global home_msg_received
    print(value)
    if not home_msg_received:
        alt_offset = value.altitude / 1000
        print("ALT_OFFSET", alt_offset)
        home_msg_received = True

def imu_msg_callback(value):
    # print(value)
    imuFile.write(f"{time.time()},{value.xacc},{value.yacc},{value.zacc},{value.xgyro},{value.ygyro},{value.zgyro},{value.xmag},{value.ymag},{value.zmag}\n")

def batt_msg_callback(value):
    global arena_batt
    arena_batt = value.voltages[0] / 4500

def mode_msg_callback(value):
    print(value)

#######################################
# Functions - T265
#######################################

def increment_reset_counter():
    global reset_counter
    if reset_counter >= 255:
        reset_counter = 1
    reset_counter += 1

# List of notification events: https://github.com/IntelRealSense/librealsense/blob/development/include/librealsense2/h/rs_types.h
# List of notification API: https://github.com/IntelRealSense/librealsense/blob/development/common/notifications.cpp
def realsense_notification_callback(notif):
    progress("INFO: T265 event: " + notif)
    if notif.get_category() is rs.notification_category.pose_relocalization:
        increment_reset_counter()

def realsense_connect():
    global pipe, pose_sensor, rs_cfg

    if (debug_enable):
        rs.log_to_console(rs.log_severity.info)
    
    # Declare RealSense pipeline, encapsulating the actual device and sensors
    pipe = rs.pipeline()

    # Build config object before requesting data
    rs_cfg = rs.config()
    ctx = rs.context()

    # Enable the stream we are interested in
    rs_cfg.enable_stream(rs.stream.pose) # Positional data

    # Configure callback for relocalization event
    device = rs_cfg.resolve(pipe).get_device()
    pose_sensor = device.first_pose_sensor()
    pose_sensor.set_option(rs.option.enable_pose_jumping, 0)
    pose_sensor.set_option(rs.option.enable_relocalization, 0)
    pose_sensor.set_option(rs.option.enable_map_preservation, 0)
    pose_sensor.set_option(rs.option.enable_mapping, 0)

    # Hard reset the sensor in an attempt to combat disconnects
    # pose_sensor.hardware_reset()

    pose_sensor.set_notifications_callback(realsense_notification_callback)

    # Start streaming with requested config
    pipe.start(rs_cfg)

#######################################
# Functions - Miscellaneous
#######################################

# Monitor user input from the terminal and perform action accordingly
def user_input_monitor():
    global scale_factor, threads_should_exit, enable_auto_set_ekf_home, home_msg_received, drone_ready, conn
    home_msg_received = False
    while not threads_should_exit and not home_msg_received:
        print('Set EKF home with default GPS location')
        set_default_global_origin()
        set_default_home_position()
        time.sleep(5) # Wait a short while for FCU to start working
    drone_ready = True

    while not threads_should_exit:
        # Add new action here according to the key pressed.
        # Enter: Set EKF home when user press enter
        try:
            c = input()
            if c == "":
                print('Set EKF home with default GPS location')
                set_default_global_origin()
                set_default_home_position()
            else:
                print("Got keyboard input %s" % c)
        except IOError: pass


#######################################
# Main code starts here
#######################################

try:
    progress("INFO: pyrealsense2 version: %s" % str(rs.__version__))
except Exception:
    # fail silently
    pass

progress("INFO: Starting Vehicle communications")
conn = mavutil.mavlink_connection(
    connection_string,
    autoreconnect = True,
    source_system = 1,
    source_component = 93,
    baud=connection_baudrate,
    force_connected=True,
)

mavlink_callbacks = {
    'ATTITUDE': att_msg_callback,
    'GLOBAL_POSITION_INT': global_msg_callback,
    'MODE':mode_msg_callback,
    'HOME_POSITION':home_msg_callback,
    'RAW_IMU':imu_msg_callback,
    'BATTERY_STATUS':batt_msg_callback
}

# connecting and configuring the camera is a little hit-and-miss.
# Start a timer and rely on a restart of the script to get it working.
# Configuring the camera appears to block all threads, so we can't do
# this internally.

signal.setitimer(signal.ITIMER_REAL, 5)  # seconds...

print('Connecting to camera...')
realsense_connect()
print('Camera connected.')

signal.setitimer(signal.ITIMER_REAL, 0)  # cancel alarm

# Try getting a frame:
signal.setitimer(signal.ITIMER_REAL, 5)  # seconds...
pipe.try_wait_for_frames(3000)
print('T265 sending frames.')
signal.setitimer(signal.ITIMER_REAL, 0)  # cancel alarm


#########################
# Signal handling
#########################

# gracefully terminate the script if an interrupt signal (e.g. ctrl-c)
# is received.  This is considered to be abnormal termination.
main_loop_should_quit = False
def sigint_handler(sig, frame):
    global main_loop_should_quit, threads_should_exit
    main_loop_should_quit = True
    threads_should_exit = True
    if (enable_arena):
        time.sleep(1)
        scene.stop_tasks()
signal.signal(signal.SIGINT, sigint_handler)

# gracefully terminate the script if a terminate signal is received
# (e.g. kill -TERM).  
def sigterm_handler(sig, frame):
    global main_loop_should_quit, threads_should_exit
    main_loop_should_quit = True
    threads_should_exit = True
    global exit_code
    exit_code = 0
    if (enable_arena):
        time.sleep(1)
        scene.stop_tasks()

signal.signal(signal.SIGTERM, sigterm_handler)


##################################
# MAIN EVENT LOOP
##################################

def main():
    print("STARTING MAIN THREAD")
    global main_loop_should_quit, pipe, H_aeroRef_aeroBody, H_aeroRef_T265Ref, H_T265body_aeroBody, V_aeroRef_aeroBody, current_confidence_level, heading_north_yaw, prev_data, pfStatus, reset_counter, realsense_resetting, current_time_us
    if compass_enabled == 1:
        time.sleep(1) # Wait a short while for yaw to be correctly initiated
    try:
        while not main_loop_should_quit:
            # Wait for the next set of frames from the camera
            status, frames = pipe.try_wait_for_frames(500)
            if not status: 
                print("--------------------------")
                print("    REALSENSE DATA BAD    ")
                print("--------------------------")
                set_mode("STABILIZE")
                pipe.stop()
                increment_reset_counter()
                realsense_resetting = True
                time.sleep(1)
                signal.setitimer(signal.ITIMER_REAL, 5)  # 5 second timeout
                realsense_connect()
                signal.setitimer(signal.ITIMER_REAL, 0)  # cancel
                realsense_resetting = False
                set_mode("GUIDED")
                continue

            # Fetch pose frame
            pose = frames.get_pose_frame()

            # Process data
            if pose:
                with drone_lock:
                    # Store the timestamp for MAVLink messages
                    current_time_us = int(round(time.time() * 1000000))

                    # Pose data consists of translation and rotation
                    data = pose.get_pose_data()
                    
                    # Confidence level value from T265: 0-3, remapped to 0 - 100: 0% - Failed / 33.3% - Low / 66.6% - Medium / 100% - High  
                    current_confidence_level = float(data.tracker_confidence * 100 / 3)  

                    # In transformations, Quaternions w+ix+jy+kz are represented as [w, x, y, z]!
                    H_T265Ref_T265body = tf.quaternion_matrix([data.rotation.w, data.rotation.x, data.rotation.y, data.rotation.z]) 
                    H_T265Ref_T265body[0][3] = data.translation.x * scale_factor
                    H_T265Ref_T265body[1][3] = data.translation.y * scale_factor
                    H_T265Ref_T265body[2][3] = data.translation.z * scale_factor

                    # Transform to aeronautic coordinates (body AND reference frame!)
                    H_aeroRef_aeroBody = H_aeroRef_T265Ref.dot( H_T265Ref_T265body.dot( H_T265body_aeroBody))

                    # Calculate GLOBAL XYZ speed (speed from T265 is already GLOBAL)
                    V_aeroRef_aeroBody = tf.quaternion_matrix([1,0,0,0])
                    V_aeroRef_aeroBody[0][3] = data.velocity.x
                    V_aeroRef_aeroBody[1][3] = data.velocity.y
                    V_aeroRef_aeroBody[2][3] = data.velocity.z
                    V_aeroRef_aeroBody = H_aeroRef_T265Ref.dot(V_aeroRef_aeroBody)

                    # Check for pose jump and increment reset_counter
                    if prev_data != None:
                        delta_translation = [data.translation.x - prev_data.translation.x, data.translation.y - prev_data.translation.y, data.translation.z - prev_data.translation.z]
                        delta_velocity = [data.velocity.x - prev_data.velocity.x, data.velocity.y - prev_data.velocity.y, data.velocity.z - prev_data.velocity.z]
                        position_displacement = np.linalg.norm(delta_translation)
                        speed_delta = np.linalg.norm(delta_velocity)

                        # Pose jump is indicated when position changes abruptly. The behavior is not well documented yet (as of librealsense 2.34.0)
                        jump_threshold = 0.1 # in meters, from trials and errors, should be relative to how frequent is the position data obtained (200Hz for the T265)
                        jump_speed_threshold = 20.0 # in m/s from trials and errors, should be relative to how frequent is the velocity data obtained (200Hz for the T265)
                        if (position_displacement > jump_threshold) or (speed_delta > jump_speed_threshold):
                            print('VISO jump detected')
                            if position_displacement > jump_threshold:
                                print("Position jumped by: %s" % position_displacement)
                            elif speed_delta > jump_speed_threshold:
                                print("Speed jumped by: %s" % speed_delta)
                            increment_reset_counter()
                        
                    prev_data = data

                    # Take offsets from body's center of gravity (or IMU) to camera's origin into account
                    if body_offset_enabled == 1:
                        H_body_camera = tf.euler_matrix(0, 0, 0, 'sxyz')
                        H_body_camera[0][3] = body_offset_x
                        H_body_camera[1][3] = body_offset_y
                        H_body_camera[2][3] = body_offset_z
                        H_camera_body = np.linalg.inv(H_body_camera)
                        H_aeroRef_aeroBody = H_body_camera.dot(H_aeroRef_aeroBody.dot(H_camera_body))

                    # Realign heading to face north using initial compass data
                    if compass_enabled == 1:
                        H_aeroRef_aeroBody = H_aeroRef_aeroBody.dot( tf.euler_matrix(0, 0, heading_north_yaw, 'sxyz'))

                    with pf_lock:
                        # print("VIO: ", np.float32(H_aeroRef_aeroBody[0][3]), np.float32(-H_aeroRef_aeroBody[1][3]), np.float32(-H_aeroRef_aeroBody[2][3]))
                        pfLocFile.write(f"VIO,{str(time.time())},{str(-H_aeroRef_aeroBody[0][3])},{str(H_aeroRef_aeroBody[1][3])},{str(-H_aeroRef_aeroBody[2][3])},0\n")
                        pf.depositVio(np.float64(time.time()), np.float32(-H_aeroRef_aeroBody[0][3]), np.float32(H_aeroRef_aeroBody[1][3]), np.float32(-H_aeroRef_aeroBody[2][3]), 0)
                        pfStatus |= 2

    except Exception as e:
        progress(e)

    except:
        print('ERROR IN SCRIPT')  
        print("Unexpected error: %s" % sys.exc_info()[0])

    finally:
        print('Closing the script...')
        # start a timer in case stopping everything nicely doesn't work.
        signal.setitimer(signal.ITIMER_REAL, 5)  # seconds...
        pipe.stop()

#################################
# Scheduler
#################################
if enable_arena:
    scene.run_once(arena_get_uwb_locations)
    scene.run_once(arena_get_target_location)
    scene.run_forever(update_arena_pos, 1000/update_arena_hz)

mavlink_thread = threading.Thread(target=mavlink_loop)
mavlink_thread.start()

if enable_uwb:
    print("STARTING UWB")
    uwb_thread = threading.Thread(target=uwb_loop)
    uwb_thread.start()

# Send MAVlink messages in the background at pre-determined frequencies
sched = BackgroundScheduler()
if enable_msg_vision_position_estimate and vision_position_estimate_msg_hz > 0:
    print("SCHEDULING POSITION EVERY: " + str(1/vision_position_estimate_msg_hz))
    sched.add_job(send_vision_position_estimate_message, 'interval', seconds = 1/vision_position_estimate_msg_hz)

if enable_msg_vision_speed_estimate and vision_speed_estimate_msg_hz > 0:
    sched.add_job(send_vision_speed_estimate_message, 'interval', seconds = 1/vision_speed_estimate_msg_hz)

# A separate thread to monitor user input
if enable_user_keyboard_input:
    user_keyboard_input_thread = threading.Thread(target=user_input_monitor)
    user_keyboard_input_thread.daemon = True
    user_keyboard_input_thread.start()
    print("INFO: Press Enter to set EKF home at default location")

main_thread = threading.Thread(target=main)
main_thread.daemon = True
main_thread.start()

sched.start()

def wait_drone_ready():
    def drone_is_ready():
        return drone_ready
    return wait_for(drone_is_ready, 1)

def fly_box():
    set_target(0, 0, -2, 0, timeout=3.5)
    time.sleep(1)
    if main_loop_should_quit:
        return

    set_target(2, 0, -2, 0, timeout=3.5)
    time.sleep(1)
    if main_loop_should_quit:
        return

    set_target(2, -2, -2, 0, timeout=3.5)
    time.sleep(1)
    if main_loop_should_quit:
        return

    set_target(0, -2, -2, 0, timeout=3.5)
    time.sleep(1)
    if main_loop_should_quit:
        return

def fly():
    fly_box()
    fly_box()
    fly_box()
    set_target(0, 0, -2, 0, timeout=3.5)
    time.sleep(1)
    land()

#Main blocking portion
wait_drone_ready()
print("START")
time.sleep(1)

set_mode("GUIDED")
set_drone_speed(0, 32)
set_drone_speed(1, 32)
set_drone_speed(2, 32)
set_drone_speed(3, 32)
time.sleep(1)
print("TAKEOFF")
takeoff(2)

box_thread = threading.Thread(target=fly)
box_thread.start()

if enable_arena:
    scene.run_tasks()
else:
    while not main_loop_should_quit:
        time.sleep(0.5)

main_loop_should_quit = True
threads_should_exit = True

#Cleanup
sched.shutdown()
mavlink_thread.join()
if enable_uwb:
    uwb_thread.join()
    uwb.close()
if enable_arena:
    scene.disconnect()
conn.close()
cameraFile.close()
arduGlobalFile.close()
arduAttFile.close()
imuFile.close()
pfOutFile.close()
pfLocFile.close()

print("INFO: Realsense pipeline and vehicle object closed.")
sys.exit(exit_code)


#!/usr/bin/env python
"""
vl53l5cx_py_basic.py: Get data from 1 VL53L5CX sensor.

Created by Perry Naseck on 11/10/21.

Copyright (c) 2021, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

import argparse

from vl53l5cx_py.driver import VL53L5CX

def main():
    parser = argparse.ArgumentParser(
        description='Get data from 1 VL53L5CX sensor.')
    parser.add_argument('dev_path', metavar='/dev/i2c-XX', type=str,
                        help='i2c device path')
    args = parser.parse_args()

    sensor = VL53L5CX(args.dev_path)
    sensor.start_ranging()

    while True:
        results = sensor.get_range()
        out = ""
        for i in range(8):
            for j in range(8):
                out += '{:4d}'.format(results["distance_mm"][(i*8) + j])
                out += " "
            out += "\n"
        print(out)

if __name__=="__main__":
   res = main()
   exit(res)



#!/usr/bin/env python
"""
licosa_py.py: Get data from a LiCosa serial device.

Created by Perry Naseck on 2/22/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

import argparse
import asyncio
from dataclasses import asdict

from aioserial import AioSerial

from licosa_py import LiCosaPacketParser

async def licosa(device="/dev/ttyAMA2", baudrate=230400):
    aioserial_instance = AioSerial(port=args.dev_path, baudrate=args.baudrate)
    pkt = LiCosaPacketParser()

    while True:
        msg_byte = await aioserial_instance.read_async(size=1)

        pkt_out = pkt.next_byte(msg_byte)
        if pkt_out is not False:
            data = asdict(pkt_out)
            print(data)




"""
__init__.py: Definitions for services.

Created by Perry Naseck on 1/25/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

from .controller import ArenaRobotServiceController
from .controller.pololu_slave_i2c import \
    ArenaRobotServiceControllerPololuSlaveI2C
from .processor import ArenaRobotServiceProcessor
from .processor.apriltag_detector import \
    ArenaRobotServiceProcessorApriltagDetector
from .processor.t265 import ArenaRobotServiceProcessorT265
from .sensor import ArenaRobotServiceSensor
from .sensor.beluga_serial import ArenaRobotServiceSensorBelugaSerial
from .sensor.licosa_serial import ArenaRobotServiceSensorLiCosaSerial
from .sensor.t265 import ArenaRobotServiceSensorT265
from .sensor.vl53l5cx import ArenaRobotServiceSensorVL53L5CX
from .service import ArenaRobotService

SERVICE_TYPE_MAP = {
    "unknown": ArenaRobotService,
    "sensor": ArenaRobotServiceSensor,
    "sensor_beluga_serial": ArenaRobotServiceSensorBelugaSerial,
    "sensor_licosa_serial": ArenaRobotServiceSensorLiCosaSerial,
    "sensor_t265": ArenaRobotServiceSensorT265,
    "sensor_vl53l5cx": ArenaRobotServiceSensorVL53L5CX,
    "processor": ArenaRobotServiceProcessor,
    "processor_t265": ArenaRobotServiceProcessorT265,
    "processor_apriltag_detector": ArenaRobotServiceProcessorApriltagDetector,
    "controller": ArenaRobotServiceController,
    "controller_pololu_slave_i2c": ArenaRobotServiceControllerPololuSlaveI2C
}



"""
parser.py: Definitions for LiCosa packet parser.

Created by Perry Naseck on 2/24/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

from enum import Enum, auto
from struct import unpack

import numpy as np

from .packets import (LICOSA_PKT_LENGTHS, LICOSA_PKT_TYPE_VALUES,
                      LiCosaErrorPacket, LiCosaIMUFullPacket, LiCosaIMUPacket,
                      LiCosaLidarPacket, LiCosaPacket, LiCosaPacketType)
from vl53l5cx_py.helpers import VL53L5CXSensorData

LICOSA_PKT_HEADER = b'\xCC'


class LiCosaState(Enum):
    """LiCosa serial sensor state class."""
    LICOSA_PKT_STATE_START = auto()
    LICOSA_PKT_STATE_TYPE = auto()
    LICOSA_PKT_STATE_DATA = auto()
    LICOSA_PKT_STATE_CHECKSUM = auto()

class LiCosaPacketParser():
    """LiCosa packet parser."""

    def __init__(self):
        """Initialize the LiCosa packet parser class."""
        self.state = LiCosaState.LICOSA_PKT_STATE_START
        self.pkt_type = None
        self.pkt_data_len = None
        self.pkt_data_index = 0
        self.pkt_buf_list = []
        self.pkt_data_buf_list = []

    @staticmethod
    def calculate_checksum(pkt_type, pkt_data_buf_list):
        """Calculate the LiCosa packet checksum."""
        return (
            (
                sum(int.from_bytes(x, 'big') for x in pkt_data_buf_list) +
                int.from_bytes(LICOSA_PKT_HEADER, 'big') +
                int.from_bytes(pkt_type.value, 'big')
            ) & 0xFF
        ).to_bytes(1, 'big')

    @staticmethod
    def parse_error_pkt_data(pkt_data_buf):
        """Parse LiCosa error packet data."""
        err_msg = int(pkt_data_buf)
        return LiCosaErrorPacket(
            data_buf=pkt_data_buf,
            err=err_msg
        )

    @staticmethod
    def parse_imu_pkt_data(pkt_data_buf):
        """Parse LiCosa IMU packet data."""
        ts, yaw, roll, pitch = unpack("<Lfff", pkt_data_buf)
        return LiCosaIMUPacket(
            data_buf=pkt_data_buf,
            ts=np.uint64(ts),
            yaw=np.float64(yaw),
            roll=np.float64(roll),
            pitch=np.float64(pitch)
        )

    @staticmethod
    def parse_imu_full_pkt_data(pkt_data_buf):
        """Parse LiCosa full IMU packet data."""
        (ts, yaw, roll, pitch, sys_calib, gyro_calib, accel_calib, mag_calib,
         sys_status, sys_error) = unpack("<LfffBBBBBBxx", pkt_data_buf)
        return LiCosaIMUFullPacket(
            data_buf=pkt_data_buf,
            ts=np.uint64(ts),
            yaw=np.float64(yaw),
            roll=np.float64(roll),
            pitch=np.float64(pitch),
            sys_calib=np.float64(sys_calib),
            gyro_calib=np.uint8(gyro_calib),
            accel_calib=np.uint8(accel_calib),
            mag_calib=np.uint8(mag_calib),
            sys_status=np.uint8(sys_status),
            sys_error=np.uint8(sys_error)
        )

    @staticmethod
    def parse_lidar_pkt_data(pkt_data_buf):
        """Parse LiCosa lidar packet data."""
        num_sensors = len(pkt_data_buf) // 128
        sensors = [None] * num_sensors
        for i in range(num_sensors):
            buf = pkt_data_buf[(i*128):((i+1)*128)]
            raw_values = unpack("<" + ("H" * 64), buf)
            distance_mm = []
            target_status = []
            for j in range(64):
                distance_mm.append(np.int16(raw_values[j]) & 0xFFF)
                target_status.append(np.uint8((np.uint16(raw_values[j]) >> 12) & 0xF))
            sensors[i] = VL53L5CXSensorData(
                distance_mm=distance_mm,
                target_status=target_status
            )

        return LiCosaLidarPacket(
            data_buf=pkt_data_buf,
            sensors=sensors
        )

    @staticmethod
    def parse_pkt_data(pkt_type, pkt_data_buf_list):
        """Parse LiCosa packet data."""
        pkt_data_buf = b''.join(pkt_data_buf_list)
        if pkt_type == LiCosaPacketType.LICOSA_PKT_TYPE_ERROR:
            return LiCosaPacketParser.parse_error_pkt_data(pkt_data_buf)
        elif pkt_type == LiCosaPacketType.LICOSA_PKT_TYPE_IMU:
            return LiCosaPacketParser.parse_imu_pkt_data(pkt_data_buf)
        elif pkt_type == LiCosaPacketType.LICOSA_PKT_TYPE_IMU_FULL:
            return LiCosaPacketParser.parse_imu_full_pkt_data(pkt_data_buf)
        elif pkt_type == LiCosaPacketType.LICOSA_PKT_TYPE_LIDAR:
            return LiCosaPacketParser.parse_lidar_pkt_data(pkt_data_buf)

        return LiCosaPacket(
            pkt_type=pkt_type,
            data_buf=pkt_data_buf
        )

    def next_byte(self, msg_byte):
        """Parse LiCosa packets."""
        self.pkt_buf_list.append(msg_byte)
        if self.state == LiCosaState.LICOSA_PKT_STATE_START:
            if msg_byte == LICOSA_PKT_HEADER:
                self.pkt_buf_list = [ msg_byte ]
                self.state = LiCosaState.LICOSA_PKT_STATE_TYPE
            else:
                print(f"LiCosa: non-start byte: {msg_byte}, bytes: {len(self.pkt_buf_list)}")

        elif self.state == LiCosaState.LICOSA_PKT_STATE_TYPE:
            if msg_byte in LICOSA_PKT_TYPE_VALUES:
                self.pkt_type = LiCosaPacketType(msg_byte)
                self.pkt_data_len = LICOSA_PKT_LENGTHS[self.pkt_type]
                self.pkt_data_index = 0
                self.pkt_data_buf_list = [None] * self.pkt_data_len
                self.state = LiCosaState.LICOSA_PKT_STATE_DATA
            else:
                print(f"LiCosa: invalid message type: {msg_byte}, bytes: {self.pkt_buf_list}")
                self.state = LiCosaState.LICOSA_PKT_STATE_START

        elif self.state == LiCosaState.LICOSA_PKT_STATE_DATA:
            self.pkt_data_buf_list[self.pkt_data_index] = msg_byte
            self.pkt_data_index += 1
            if self.pkt_data_index == self.pkt_data_len:
                self.state = LiCosaState.LICOSA_PKT_STATE_CHECKSUM

        elif self.state == LiCosaState.LICOSA_PKT_STATE_CHECKSUM:
            msg_chk_cal = self.calculate_checksum(self.pkt_type, self.pkt_data_buf_list)
            if msg_chk_cal != msg_byte:
                print(f"LiCosa: invalid checksum! type: {self.pkt_type}, computed: {msg_chk_cal}, received: {msg_byte}, bytes: {self.pkt_buf_list}")
                self.state = LiCosaState.LICOSA_PKT_STATE_START
            else:
                self.state = LiCosaState.LICOSA_PKT_STATE_START
                return self.parse_pkt_data(self.pkt_type, self.pkt_data_buf_list)

        return False



"""
packets.py: Definitions for LiCosa packets and helpers.

Created by Perry Naseck on 2/24/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

from enum import Enum
from dataclasses import dataclass
from json import JSONEncoder
from typing import ClassVar, List

import numpy as np

from vl53l5cx_py.helpers import VL53L5CXSensorData

class LiCosaPacketType(Enum):
    """LiCosa serial sensor packet type class."""
    LICOSA_PKT_TYPE_ERROR = b'\x01'
    LICOSA_PKT_TYPE_IMU = b'\x02'
    LICOSA_PKT_TYPE_IMU_FULL = b'\x03'
    LICOSA_PKT_TYPE_LIDAR = b'\x04'

LICOSA_PKT_TYPE_VALUES = set(item.value for item in LiCosaPacketType)

# Excludes start/type/checksum
LICOSA_PKT_LENGTHS = {
    LiCosaPacketType.LICOSA_PKT_TYPE_ERROR: 1,
    LiCosaPacketType.LICOSA_PKT_TYPE_IMU: 16,

    # 22 bytes of data, 2 bytes of padding as per C struct-padding
    LiCosaPacketType.LICOSA_PKT_TYPE_IMU_FULL: 24,
    LiCosaPacketType.LICOSA_PKT_TYPE_LIDAR: 640
}

@dataclass
class _LiCosaPacketBase:
    """Base dataclass for LiCosa packets."""
    data_buf: bytes


@dataclass
class LiCosaPacket(_LiCosaPacketBase):
    """Dataclass for LiCosa packets."""
    pkt_type: LiCosaPacketType


@dataclass
class _LiCosaErrorPacketBase(_LiCosaPacketBase):
    """Error LiCosa packet base."""
    err: int


@dataclass
class LiCosaErrorPacket(LiCosaPacket, _LiCosaErrorPacketBase):
    """Error LiCosa packet."""
    PKT_TYPE: ClassVar[LiCosaPacketType] = LiCosaPacketType.LICOSA_PKT_TYPE_ERROR
    pkt_type: LiCosaPacketType = LiCosaPacketType.LICOSA_PKT_TYPE_ERROR


@dataclass
class _LiCosaIMUPacketBase(_LiCosaPacketBase):
    """IMU LiCosa packet base."""
    ts: np.uint64
    yaw: np.float64
    roll: np.float64
    pitch: np.float64


@dataclass
class LiCosaIMUPacket(LiCosaPacket, _LiCosaIMUPacketBase):
    """IMU LiCosa packet."""
    PKT_TYPE: ClassVar[LiCosaPacketType] = LiCosaPacketType.LICOSA_PKT_TYPE_IMU
    pkt_type: LiCosaPacketType = LiCosaPacketType.LICOSA_PKT_TYPE_IMU


@dataclass
class _LiCosaIMUFullPacketBase(_LiCosaIMUPacketBase):
    """Full IMU LiCosa packet base."""
    sys_calib: np.uint8
    gyro_calib: np.uint8
    accel_calib: np.uint8
    mag_calib: np.uint8
    sys_status: np.uint8
    sys_error: np.uint8


@dataclass
class LiCosaIMUFullPacket(LiCosaIMUPacket, _LiCosaIMUFullPacketBase):
    """Full IMU LiCosa packet."""
    PKT_TYPE: ClassVar[LiCosaPacketType] = LiCosaPacketType.LICOSA_PKT_TYPE_IMU_FULL
    pkt_type: LiCosaPacketType = LiCosaPacketType.LICOSA_PKT_TYPE_IMU_FULL


@dataclass
class _LiCosaLidarPacketBase(_LiCosaPacketBase):
    """Lidar LiCosa packet base."""
    sensors: List[VL53L5CXSensorData]


@dataclass
class LiCosaLidarPacket(LiCosaPacket, _LiCosaLidarPacketBase):
    """Lidar LiCosa packet."""
    PKT_TYPE: ClassVar[LiCosaPacketType] = LiCosaPacketType.LICOSA_PKT_TYPE_LIDAR
    pkt_type: LiCosaPacketType = LiCosaPacketType.LICOSA_PKT_TYPE_LIDAR

class LiCosaJSONEncoder(JSONEncoder):
    """JSON Encoder helper for LiCosa packets."""

    def default(self, obj):
        """JSON Encoder helper function for LiCosa packets."""
        if isinstance(obj, (np.int_, np.int16, np.uint8, np.uint16, np.uint64)):
            return int(obj)

        elif isinstance(obj, (np.float_, np.float64)):
            return float(obj)

        elif isinstance(obj, (LiCosaPacketType)):
            return obj.name

        return JSONEncoder.default(self, obj)



#!/usr/bin/env python3
from enum import Enum, auto
from arena import *
import time
import sys

scene = Scene(host="mqtt.arenaxr.org", scene="ARENA-drone", realm="realm", namespace="pnaseck")

HEIGHT = 1.5

@scene.run_once
def arena_init():
    global drone_target, pushpin, drone_target_ind

    try:
        drone_target = scene.get_persisted_obj("drone_target")
        assert drone_target is not None
        assert len(sys.argv) == 1 # ONLY USE PERSIST IF NO ARGUMENTS GIVEN
    except:
        drone_target = Box(
            object_id="drone_target",
            position=(0,HEIGHT,0),
            scale=(0.1,0.1,0.1),
            color=(204,0,0),
            clickable=True,
            persist=True
        )

    pushpin = Cylinder(
        object_id="pushpin",
        position=(999, 999, 999),
        rotation=(0, 0, 0),
        scale=(.1, .02, .1),
        color=(0, 120, 0),
        persist=True,
        clickable=True,
        evt_handler=click_handler
    )

    drone_target_ind = Cylinder(
        object_id="drone_target_ind",
        position=(999, 999, 999),
        rotation=(0, 0, 0),
        scale=(.1, HEIGHT*0.95, .1),
        material=Material(color="#007000", transparent=True, opacity=0.5),
        persist=True
    )

    origin = Box(
        object_id="origin",
        position=(0, 0, 0),
        scale=(.05, .05, .05),
        color=(120, 120, 120),
        persist=True,
        clickable=True,
        evt_handler=click_handler
    )

    scene.add_object(drone_target)
    scene.add_object(drone_target_ind)
    scene.add_object(pushpin)
    scene.add_object(origin)

    scene.on_msg_callback = on_message


done = False
landed = False
land_time = 0

def click_handler(scene, evt, msg):
    global done, land_time
    if done: return

    if evt.type == "mousedown":
        pos = pushpin.data.position

        if abs(pos["x"]) < 0.4 and abs(pos["z"]) < 0.4:
            wp = (0, HEIGHT, 0)
            drone_target_ind.update_attributes(position=Position(0, HEIGHT*0.5, 0), persist=True)
            drone_target_ind.data.color = Color(220,40,40)
            scene.update_object(drone_target_ind)

            land_time = time.time()
            done = True
            pushpin.data.position = Position(999, 999, 999)
            scene.update_object(pushpin)
        else:
            wp = (pos["x"], HEIGHT, pos["z"])
            drone_target_ind.update_attributes(position=Position(wp[0], HEIGHT*0.5, wp[2]), persist=True)
            scene.update_object(drone_target_ind)

        drone_target.update_attributes(position=Position(*wp), persist=True)
        scene.update_object(drone_target)

def quaternion_mult(q,r):
    return [r[0]*q[0]-r[1]*q[1]-r[2]*q[2]-r[3]*q[3],
            r[0]*q[1]+r[1]*q[0]-r[2]*q[3]+r[3]*q[2],
            r[0]*q[2]+r[1]*q[3]+r[2]*q[0]-r[3]*q[1],
            r[0]*q[3]-r[1]*q[2]+r[2]*q[1]+r[3]*q[0]]

def point_rotation_by_quaternion(point,q):
    r = [0]+point
    q_conj = [q[0],-1*q[1],-1*q[2],-1*q[3]]
    return quaternion_mult(quaternion_mult(q,r),q_conj)[1:]

def on_message(scene, evt, msg):
    try:
        oid = msg.get("object_id")

        global landed, land_time

        if landed:
            if time.time() - land_time > 6:
                sys.exit(0)
            return

        if done:
            if time.time() - land_time > 6:
                drone_target.data.position = Position(0, 0, 0)
                scene.update_object(drone_target)
                land_time = time.time()
                landed = True

                drone_target_ind.update_attributes(position=Position(999, 999, 999), persist=True)
                scene.update_object(drone_target_ind)

            return

        if not oid.startswith("camera"): return
        # Get the position and orientation of the user's head
        pos = msg["data"]["position"]
        rot = msg["data"]["rotation"]
        rot = [rot["w"], rot["x"], rot["y"], rot["z"]]

        # Compute a raycast from the user's head to the ground plane
        vec = point_rotation_by_quaternion([0, 0, -1], rot)
        res = (0 - pos["y"]) / vec[1]
        x = pos["x"] + res*vec[0]
        z = pos["z"] + res*vec[2]

        if abs(x) < 0.4 and abs(z) < 0.4:
            pushpin.data.color = Color(220,40,40)
        else:
            pushpin.data.color = Color(63,150,210)

        pushpin.data.position = Position(x, 0.01, z)
        scene.update_object(pushpin)
    except:
        pass

scene.run_tasks()




"""
const.py: Constants for the occupancy grid.

Created by Perry Naseck on 6/29/21.

Copyright (c) 2021, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

UPDATE_ACTION = 'action'
UPDATE_LOC = 'loc'
UPDATE_OCCUPANT = 'occupant'

UPDATE_ACTION_OCCUPY = 'occupy'
UPDATE_ACTION_LEAVE = 'leave'



import numpy as np
import serial
import time
from particlefilter import ParticleFilterLoc

pf = ParticleFilterLoc()
uwb = serial.Serial('/dev/ttyAMA1', 115200, timeout=1)

uwb_locations = {
  7: [1.828, 1.749, 2.558],
  8: [2.806, 0.799, -1.901],
  9: [-1.82, 0.168, 1.955],
  10: [-2.94, 2.804, -2.126],
  11: [1.334, 2.052, -3.678]
}

uwb.write(b"AT+RESET 20\n")
time.sleep(1)
uwb.write(b"AT+ID 25\n")
time.sleep(1)
uwb.write(b"AT+RATE 20\n")
time.sleep(1)
uwb.write(b"AT+STARTBLE\n")
time.sleep(1)
uwb.write(b"AT+STARTUWB\n")
time.sleep(1)
uwb.write(b"AT+STREAMMODE 0\n")
time.sleep(1)

uwb.readline()
count = 0
startTime = time.time()
while True:
  line = uwb.readline()
  print(line)
  if len(line) > 0 and (line[0] != ord('#')):
    line = line[0:-2].replace(b' ', b'').decode('utf-8')
    #ID, RANGE, RSSI, TIMESTAMP
    curData = line.split(',')
    if (len(curData) < 4):
      print(curData)
      continue
    if int(curData[0]) not in uwb_locations:
      continue
    x, y, z = uwb_locations[int(curData[0])]
    # print (x, y, z, float(curData[1]))
    curTime = time.time()
    # pf.depositVio(np.float64(curTime), 0, 0, 0, np.float32(0.5*(curTime - startTime)))

    pf.depositRange(np.float32(x), np.float32(y), np.float32(z), np.float32(float(curData[1])), 0.1)

    count += 1

    pf_data = pf.getTagLoc()
    # box.update_attributes(position=Position(x.value, y.value, z.value), rotation=Rotation(0, theta.value*RAD_TO_DEG, 0))
    # scene.update_object(box)
    print(pf_data[2],"\t", pf_data[3],"\t", pf_data[4])



"""
pololu_slave_i2c.py: Interface for Pololu Slave I2C devices.

Created by Perry Naseck on 6/30/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

# This is essentially a port of the Pololu code from smbus to periphery with
# dynamic buffer format configuration:
# https://github.com/pololu/pololu-rpi-slave-arduino-library/blob/master/pi/a_star.py

from struct import calcsize, pack, unpack
from time import sleep
from typing import Sequence

from paho.mqtt.client import MQTTMessage
from periphery import I2C, I2CError

from arenarobot.service.controller import ArenaRobotServiceController


# pylint: disable=too-many-instance-attributes
class ArenaRobotServiceControllerPololuSlaveI2C(ArenaRobotServiceController):
    """Pololu Slave I2C device class for the ARENA."""

    DEVICE_INSTANCE_CONTROLLER_TYPE = "pololu_slave_i2c"

    # pylint: disable=too-many-branches
    def __init__(self, dev_path: str,
                 slave_data_types: Sequence[dict],
                 slave_addr: int = 20,
                 **kwargs):
        """Initialize the Pololu Slave I2C device controller class."""
        self.dev_path = dev_path
        self.slave_addr = slave_addr
        self.slave_data_types = slave_data_types

        for name, data_type in self.slave_data_types.items():
            if len(name) < 1:
                raise ValueError("Data type names must not be blank!")
            if "format" not in data_type:
                raise ValueError("Data type missing format char!")
            if not isinstance(data_type["format"], str):
                raise ValueError("Data type format not a str!")
            if len(data_type["format"]) != 1:
                raise ValueError("Data type format option should be one char!")
            if "num_items" not in data_type:
                raise ValueError("Data type missing num_items!")
            if (not isinstance(data_type["num_items"], int) or
                    data_type["num_items"] < 1):
                raise ValueError("Data type num_items must be int at least 1!")
            if "write" not in data_type:
                raise ValueError("Data type missing write!")
            if not isinstance(data_type["write"], bool):
                raise ValueError("Data type write not a bool!")
            if data_type["write"]:
                if "initial" not in data_type:
                    raise ValueError("Writable data type missing initial!")
                if data_type["format"] == "s":
                    if not isinstance(data_type["initial"], str):
                        raise ValueError("Data type initial string not a str!")
                    intitial_bytes = bytes(data_type["initial"], "utf-8")
                    if len(intitial_bytes) != data_type["num_items"]:
                        raise ValueError("Data type initial wrong length!")
                elif len(data_type["initial"]) != data_type["num_items"]:
                    if not isinstance(data_type["initial"], list):
                        raise ValueError("Data type initial list not a list!")
                    raise ValueError("Data type initial wrong length!")

        self.data_types_write = []
        self.i2c_bus = None
        self.struct_format_str = "<"  # Little endian
        self.struct_size = 0

        controller_type = (ArenaRobotServiceControllerPololuSlaveI2C
                           .DEVICE_INSTANCE_CONTROLLER_TYPE)
        super().__init__(
            device_instance_controller_type=controller_type,
            device_instance_prefix="pololu_slave_i2c_",
            **kwargs
        )

    def setup(self):
        """Set up Pololu Slave I2C controller."""
        for name, data_type in self.slave_data_types.items():
            if data_type["write"]:
                self.data_types_write.append(name)
            data_type["addr"] = self.struct_size
            data_type["full_format"] = (
                str(data_type["num_items"]) + data_type["format"])
            data_type["size"] = calcsize(data_type["format"])
            data_type["full_size"] = calcsize(data_type["full_format"])
            self.struct_size += data_type["full_size"]
            self.struct_format_str += data_type["full_format"]

        self.i2c_bus = I2C(self.dev_path)

        # write start byte and read and ignore first message
        self.i2c_bus.transfer(self.slave_addr, [I2C.Message([0], read=False)])
        sleep(0.0001)
        self.i2c_bus.transfer(
            self.slave_addr,
            [I2C.Message([0x0]*self.struct_size, read=True)])
        sleep(0.0001)

        # reset initials
        for name in self.data_types_write:
            initial = self.slave_data_types[name]["initial"]
            self.write_val(name, initial)
            sleep(0.0001)

        super().setup()

    def write_val(self, param_name: str, value, index: int = 0):
        """Write value to Pololu Slave I2C device."""
        if param_name not in self.data_types_write:
            raise ValueError("Param not found or not in write list!")
        data_type = self.slave_data_types[param_name]
        addr = data_type["addr"]
        format_char = data_type["format"]
        packed_format = format_char
        num_items = data_type["num_items"]

        if index > 0 and format_char == "s":
            raise ValueError("Indices cannot be used for strings!")
        if index > 0 and num_items <= 1:
            raise ValueError("Index too high for this data type!")
        if format_char == "s" and not isinstance(value, str):
            raise ValueError("String value must be a string!")
        if isinstance(value, Sequence) and len(value) > num_items - index:
            raise ValueError("Too many items in value list!")
        if isinstance(value, Sequence) and len(value) == 0:
            raise ValueError("Value list is empty!")

        addr += data_type["size"] * index
        if format_char == "s":
            packed_format = data_type["full_format"]
            value = (bytes(value, "utf-8"),)
        elif not isinstance(value, Sequence):
            value = (value,)
        else:
            packed_format = data_type["format"] * len(value)

        packed = list(pack(packed_format, *value))
        msgs = [I2C.Message([addr, *packed])]
        self.i2c_bus.transfer(self.slave_addr, msgs)

    def fetch(self):
        """Fetch Pololu Slave I2C data."""
        # Set up for read
        msgs = [I2C.Message([0x0], read=False)]
        try:
            self.i2c_bus.transfer(self.slave_addr, msgs)
        except I2CError as err:
            print(err)
            self.publish({"data": {"error": str(err)}})
            print("continuing")
            return

        # Required delay, see Pololu a_star.py for explanation. This was
        # to also be required in this implementation. Otherwise, the first
        # few fields would be unintentionally written to.
        sleep(0.0001)

        # Read
        msgs = [I2C.Message([0x0]*self.struct_size, read=True)]
        try:
            self.i2c_bus.transfer(self.slave_addr, msgs)
        except I2CError as err:
            print(err)
            self.publish({"data": {"error": str(err)}})
            print("continuing")
            return
        buffer_in = unpack(self.struct_format_str, bytes(msgs[0].data))
        data = {}
        index = 0
        for name in self.slave_data_types:
            num_items = self.slave_data_types[name]["num_items"]
            format_char = self.slave_data_types[name]["format"]
            if format_char == "s":  # bytes from a string are special in struct
                data[name] = buffer_in[index].decode("utf-8")
                index += 1
            elif num_items == 1:
                data[name] = buffer_in[index]
                index += num_items
            else:
                data[name] = []
                for count in range(num_items):
                    data[name].append(buffer_in[index + count])
                index += num_items
        self.publish({"data": {"pololu_device": data}})

    # pylint: disable=unused-argument
    def msg_rx(self, client, userdata, msg: MQTTMessage):
        """Receive messages to controll the Pololu I2C device."""
        payload = self.decode_payload(msg)
        if payload is None:
            return

        # Only receive commands destined for this instance
        if ("target_device_instance_name" not in payload or
                payload["target_device_instance_name"] != self.instance_name):
            return
        if "msg" not in payload:
            print("Skipping missing msg for",
                  payload["target_device_instance_name"],
                  payload)
            return
        if "data" not in payload["msg"]:
            print("Skipping missing data for",
                  payload["target_device_instance_name"],
                  payload["msg"])
            return
        data = payload["msg"]["data"]

        if "pololu_device_commands" not in data:
            print("Skipping missing commands for",
                  payload["target_device_instance_name"],
                  payload["msg"])
            return

        for command in data["pololu_device_commands"]:
            try:
                name = command["name"]
                value = command["value"]
                if "index" in command:
                    self.write_val(name, value, command["index"])
                else:
                    self.write_val(name, value)
            except ValueError as err:
                print(err)
                self.publish({"data": {"error": str(err)}})
            except Exception as err:  # pylint: disable=broad-except
                print(err)
                self.publish({"data": {"error": str(err)}})
                print("continuing")
            sleep(0.0001)



"""
t265.py: Definition for Intel T265 sensors.

Created by Perry Naseck on 5/6/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

import time
from json import JSONEncoder, dumps, loads

from arenarobot.service.sensor import ArenaRobotServiceSensor

# Optional imports allow for constants from this module to be referenced
# and for the service to be recognized without all of the required
# dependencies installed need for starting this service.
imports_missing = []
try:
    import pyrealsense2.pyrealsense2 as rs
except ImportError as err:
    rs = err
    imports_missing.append(err)


# pylint: disable=too-many-instance-attributes
class ArenaRobotServiceSensorT265(ArenaRobotServiceSensor):
    """T265 sensor class for the ARENA."""

    DEVICE_INSTANCE_SENSOR_TYPE = "t265"

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-statements
    def __init__(self, timeout_ms: int = 500,
                 enable_pose_jumping: bool = False,
                 enable_relocalization: bool = False,
                 enable_map_preservation: bool = False,
                 enable_mapping: bool = False,
                 enable_dynamic_calibration: bool = True, **kwargs):
        """Initialize the T265 sensor class."""
        # Check for missing imports
        if len(imports_missing) > 0:
            raise Exception(imports_missing)

        self.timeout_ms = timeout_ms
        self.enable_pose_jumping = enable_pose_jumping
        self.enable_relocalization = enable_relocalization
        self.enable_map_preservation = enable_map_preservation
        self.enable_mapping = enable_mapping
        self.enable_dynamic_calibration = enable_dynamic_calibration

        self.rs_pipe = None
        self.rs_cfg = None
        self.rs_device = None
        self.pose_sensor = None
        self.rs_ctx = None

        sensor_type = (ArenaRobotServiceSensorT265
                       .DEVICE_INSTANCE_SENSOR_TYPE)
        super().__init__(
            device_instance_sensor_type=sensor_type,
            device_instance_prefix="t265_",
            **kwargs
        )

    def setup(self):
        """Set up T265 sensor."""
        rs.log_to_console(rs.log_severity.warn)
        self.rs_pipe = rs.pipeline()
        self.rs_cfg = rs.config()
        self.rs_cfg.enable_stream(rs.stream.pose)

        self.rs_device = self.rs_cfg.resolve(self.rs_pipe).get_device()
        self.pose_sensor = self.rs_device.first_pose_sensor()
        self.pose_sensor.set_option(rs.option.enable_pose_jumping,
                                    int(self.enable_pose_jumping))
        self.pose_sensor.set_option(rs.option.enable_relocalization,
                                    int(self.enable_relocalization))
        self.pose_sensor.set_option(rs.option.enable_map_preservation,
                                    int(self.enable_map_preservation))
        self.pose_sensor.set_option(rs.option.enable_mapping,
                                    int(self.enable_mapping))
        self.pose_sensor.set_option(rs.option.enable_dynamic_calibration,
                                    int(self.enable_dynamic_calibration))

        self.rs_ctx = rs.context()

        def rs_notification_callback(notif):
            print("INFO: T265 event: " + notif)
            self.publish({"data": {"notif": notif}})

        self.pose_sensor.set_notifications_callback(rs_notification_callback)
        self.rs_pipe.start(self.rs_cfg)

        time.sleep(1)
        super().setup()

    def fetch(self):
        """Fetch T265 data."""
        status, frames = self.rs_pipe.try_wait_for_frames(self.timeout_ms)
        if not status:
            print("REALSENSE DATA BAD!")
            self.publish({"data": {"error": "rs_data_bad"}})
            return

        # Fetch pose frame
        pose = frames.get_pose_frame()
        if not pose:
            print("No pose!")
            self.publish({"data": {"error": "rs_no_pose"}})
            return

        pose_data = pose.get_pose_data()
        serializable_pose_data = {
            "acceleration": pose_data.acceleration,
            "angular_acceleration": pose_data.angular_acceleration,
            "angular_velocity": pose_data.angular_velocity,
            "mapper_confidence": pose_data.mapper_confidence,
            "rotation": pose_data.rotation,
            "tracker_confidence": pose_data.tracker_confidence,
            "translation": pose_data.translation,
            "velocity": pose_data.velocity
        }
        serializable_pose_data = loads(dumps(
            serializable_pose_data,
            cls=T265PoseJSONEncoder
        ))

        self.publish({"data": {"pose": serializable_pose_data}})


class T265PoseJSONEncoder(JSONEncoder):
    """JSON Encoder helper for T265 pose attributes."""

    def default(self, o):
        """JSON Encoder helper function for T265 pose attributes."""
        if isinstance(o, rs.vector):
            return {"x": o.x, "y": o.y, "z": o.z}

        if isinstance(o, rs.quaternion):
            return {"w": o.w, "x": o.x, "y": o.y, "z": o.z}

        return JSONEncoder.default(self, o)



"""
__init__.py: Definitions for controllers.

Created by Perry Naseck on 7/26/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

from arenarobot.service.service import ArenaRobotService


class ArenaRobotServiceController(ArenaRobotService):
    """Controller class for the ARENA."""

    DEVICE_INSTANCE_TYPE = "controller"
    DEVICE_INSTANCE_MULTI_TYPE = "unknown"

    def __init__(self, device_instance_prefix="",
                 device_instance_controller_type="unknown",
                 subtopic="",
                 **kwargs):
        """Initialize the controller class."""
        self.device_instance_controller_type = device_instance_controller_type
        device_instance_type = ArenaRobotServiceController.DEVICE_INSTANCE_TYPE
        super().__init__(
            device_instance_type=device_instance_type,
            device_instance_prefix=f"controller_{device_instance_prefix}",
            subtopic=f"controllers/{subtopic}",
            **kwargs
        )

    def publish_msg_base(self):
        """Publish message base."""
        payload = super().publish_msg_base()
        payload.update({
            "device_instance_controller_type":
                self.device_instance_controller_type
        })
        return payload



"""Package version information."""
# coding: utf-8
# file generated by setuptools_scm
# don't change, don't track in version control
# flake8: noqa
version = '2.2.0'  # pylint: disable=invalid-name
version_tuple = (2, 2, 0)  # pylint: disable=invalid-name



"""
transform.py: Definitions for LiCosa packet transformer.

Created by Perry Naseck on 2/24/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""


class LiCosaPacketTransformer():
    """LiCosa packet transformer."""



#!/usr/bin/env python
"""
vl53l5cx_py_basic.py: Get data from multiple VL53L5CX sensors.

Created by Perry Naseck on 12/7/21.

Copyright (c) 2021, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

import argparse
import numpy as np
import time

from periphery import GPIO
from vl53l5cx_py.driver import VL53L5CX

def main():
    parser = argparse.ArgumentParser(
        description='Get data from 1 VL53L5CX sensor.')
    parser.add_argument('dev_path', metavar='/dev/i2c-XX', type=str,
                        help='i2c device path')
    parser.add_argument('gpio_path', metavar='/dev/gpiochipXX', type=str,
                        help='gpio device path')
    parser.add_argument('rst_pin', type=int, help='i2c rst pin')
    parser.add_argument('sensor_pins', type=int, nargs='+',
                        help='LPn GPIO pins')
    args = parser.parse_args()

    sensors_data = []
    sensors = [None] * len(args.sensor_pins)

    rst_pin = GPIO(args.gpio_path, args.rst_pin, "out")
    rst_pin.write(False)

    for i in range(len(args.sensor_pins)):
        pin_num = args.sensor_pins[i]
        pin = GPIO(args.gpio_path, pin_num, "out")
        pin.write(False)
        addr = VL53L5CX.DEFAULT_ADDR + np.uint16(2 * i)
        sensors_data.append((i, pin, addr,))
        print(f'Will setup sensor {i}, addr {addr:02x} on pin {pin_num}')

    rst_pin.write(True)
    time.sleep(1)
    rst_pin.write(False)
    time.sleep(1)

    for data in sensors_data:
        i, pin, addr = data
        pin.write(True)
        time.sleep(1)
        sensors[i] = VL53L5CX(args.dev_path, target_addr=addr)
        pin.write(False)
        time.sleep(1)

    for data in sensors_data:
        i, pin, addr = data
        pin.write(True)
        time.sleep(1)
        sensors[i].start_ranging()

    while True:
        out = ""
        for i in range(len(sensors)):
            results = sensors[i].get_range()
            out += f"{i}:\n"
            for i in range(8):
                for j in range(8):
                    out += '{:4d}'.format(results["distance_mm"][(i*8) + j])
                    out += " "
                out += "\n"
        print(out)

if __name__=="__main__":
   res = main()
   exit(res)



#!/usr/bin/env python3
from enum import Enum, auto

from arena import *

scene = Scene(host="mqtt.arenaxr.org", scene="ARENA-drone", realm="realm", namespace="pnaseck")

HEIGHT = 1.5
positions = [
    [1.5, HEIGHT, 0],
    [1.5, HEIGHT, -1.5],
    [0, HEIGHT, -1.5],
    [0, HEIGHT, 0],
]
current_pos = 0

class Mode(Enum):
    INIT = auto()
    FLY = auto()
    GOTO_LAND = auto()
    LAND = auto()
    DELETE = auto()
    DONE = auto()
mode = Mode.INIT

added = False

drone_target = Box(
    object_id="drone_target",
    position=(0,HEIGHT,0),
    scale=(0.1,0.1,0.1),
    color=(204,0,0),
    clickable=True,
    persist=True
)

def click(scene, evt, msg):
    global mode
    next_mode = mode
    if evt["type"] == "mousedown":
        if mode == Mode.INIT:
            next_mode = Mode.FLY
        elif mode == Mode.FLY:
            next_mode = Mode.GOTO_LAND
    mode = next_mode

@scene.run_forever(interval_ms=4000)
def move_target():
    global current_pos, positions, mode, HEIGHT, added 
    print(mode.name)

    next_mode = mode
    if mode == Mode.INIT:
        pos = [0, HEIGHT, 0]
        scene.update_object(drone_target,persist=True)
        next_mode = Mode.INIT
    elif mode == Mode.FLY:
        pos = positions[current_pos]
        current_pos += 1
        if current_pos >= len(positions):
            current_pos = 0
        next_mode = Mode.FLY
    elif mode == Mode.GOTO_LAND:
        pos = [0, HEIGHT, 0]
        next_mode = Mode.LAND
    elif mode == Mode.LAND:
        pos = [0, 0, 0]
        next_mode = Mode.DELETE
    elif mode == Mode.DELETE:
        next_mode = Mode.DONE
    elif mode == Mode.DONE:
        next_mode = Mode.DONE
    else:
        raise Exception("Unknown mode", mode)

    if not added: 
        scene.add_object(drone_target)
        added = True
        
    if mode == Mode.DELETE:
        scene.delete_object(drone_target)
    elif mode != Mode.DONE:
        drone_target.update_attributes(position=Position(pos[0], pos[1], pos[2]), persist=True)
        scene.update_object(drone_target, clickable=True, evt_handler=click)

    mode = next_mode

scene.run_tasks()



"""
apriltag_detector.py: reads frames, detects apriltags, and returns pose.

Created by John Pedraza on 6/7/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

from json import JSONEncoder, dumps, loads
from typing import List

import numpy as np

from arenarobot.service.processor import ArenaRobotServiceProcessor

# Optional imports allow for constants from this module to be referenced
# and for the service to be recognized without all of the required
# dependencies installed need for starting this service.
imports_missing = []
try:
    import cv2
except ImportError as err:
    cv2 = err
    imports_missing.append(err)
try:
    from dt_apriltags import Detector
except ImportError as err:
    Detector = err
    imports_missing.append(err)

OPENCV_RES_INDEX_HORIZ = 3
OPENCV_RES_INDEX_VERT = 4

FLIP_MATRIX = np.array([[1,  0,  0, 0],
                        [0, -1,  0, 0],
                        [0,  0, -1, 0],
                        [0,  0,  0, 1]], dtype=float)


def detection2matrix(translation, rotation):
    """
    Reformat detected translation and rotation into 4x4 matrix.

    upper-left 3x3 matrix is the rotation matrix.
    rightmost column is translation vector (x, y, z, top to bottom).
    bottom row is [0, 0, 0, 1] to help with transformations.
    """
    translation = np.array([[t[0] for t in translation]])
    temp_mtx = np.concatenate((rotation, translation.T),
                              axis=1)
    temp_mtx = np.concatenate((temp_mtx, [np.array([0, 0, 0, 1])]),
                              axis=0)
    return temp_mtx


# pylint: disable=too-many-instance-attributes
class ArenaRobotServiceProcessorApriltagDetector(ArenaRobotServiceProcessor):
    """Apriltag Detector processor class for the ARENA."""

    DEVICE_INSTANCE_PROCESSOR_TYPE = "apriltag_detector"

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-statements
    def __init__(self,
                 video_file: str,
                 camera_resolution: List[int],
                 camera_params: List[int],
                 dist_params: List[int],
                 apriltag_locations: dict,
                 apriltag_family: str = 'tag36h11',
                 num_detector_threads: int = 1,
                 quad_decimate: float = 2.0,
                 quad_sigma: float = 0.0,
                 refine_edges: int = 1,
                 decode_sharpening: float = 0.25,
                 tag_size: float = 0.15,
                 **kwargs):
        """Initialize the apriltag detector processor class."""
        # Check for missing imports
        if len(imports_missing) > 0:
            raise Exception(imports_missing)

        # video_capture
        self.cap = video_file

        # camera resolution [width, height] in pixels
        self.camera_resolution = camera_resolution

        # camera matrix parameters fx, fy, cx, cy
        self.params = camera_params

        # distortion parameters k1, k2, p1, p2, k3
        self.dist_params = np.array(dist_params)

        # known apriltag poses (world coordinates in meters)
        self.apriltags = apriltag_locations

        # documentation for below parameters can be found at
        # https://github.com/duckietown/lib-dt-apriltags
        self.apriltag_family = apriltag_family
        self.num_detector_threads = num_detector_threads
        self.quad_decimate = quad_decimate
        self.quad_sigma = quad_sigma
        self.refine_edges = refine_edges
        self.decode_sharpening = decode_sharpening
        self.tag_size = tag_size

        self.at_detector = None
        self.mtx = None

        processor_type = (ArenaRobotServiceProcessorApriltagDetector
                          .DEVICE_INSTANCE_PROCESSOR_TYPE)

        super().__init__(
            device_instance_processor_type=processor_type,
            device_instance_prefix="apriltag_detector_",
            **kwargs
        )

    def setup(self):
        """Set up video capture and Apriltag Detector object."""
        self.cap = cv2.VideoCapture(self.cap)

        # subsample to increasing processing speed
        # note: camera parameters depend on which resolution you're using
        self.cap.set(OPENCV_RES_INDEX_HORIZ, self.camera_resolution[0])
        self.cap.set(OPENCV_RES_INDEX_VERT, self.camera_resolution[1])

        # camera matrix
        self.mtx = np.array([[self.params[0], 0.000000000000, self.params[2]],
                             [0.000000000000, self.params[1], self.params[3]],
                             [0.000000000000, 0.000000000000, 1.000000000000]])

        # apriltag detector
        # set quad_decimate to 2.0 for faster detections, worse accuracy
        # setting debug=1 saves images - don't use, runs very slowly
        self.at_detector = Detector(searchpath=['apriltags'],
                                    families=self.apriltag_family,
                                    nthreads=self.num_detector_threads,
                                    quad_decimate=self.quad_decimate,
                                    quad_sigma=self.quad_sigma,
                                    refine_edges=self.refine_edges,
                                    decode_sharpening=self.decode_sharpening,
                                    debug=0)

        super().setup()

    # Planned: figure out how this works with different users/devices
    #     (not just jpedraza/john-pi)
    #     - so the device for the topic is just specified when
    #       running arena-robot-service
    # Planned: handle getting wrong representation
    #     (apriltag solver has two possible solutions)
    # Planned: handle multiple apriltags in frame at same time
    # Planned: cap.release()?
    def fetch(self):
        """
        Capture a camera frame, detect apriltags, and calculate pose.

        Publish pose information to MQTT topic:
            realm/d/<user>/<device>/processors/apriltag_pose
        """
        # get grayscale frame for apriltag detection
        ret, frame = self.cap.read()
        if not ret:
            print('read() failed')
            return

        gray_image = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

        # undistort frame
        height, width = gray_image.shape[:2]
        new_camera_mtx, _ = cv2.getOptimalNewCameraMatrix(self.mtx,
                                                          self.dist_params,
                                                          (width, height), 1,
                                                          (width, height))
        dst = cv2.undistort(gray_image, self.mtx,
                            self.dist_params, None, new_camera_mtx)

        # detect apriltags
        tags = self.at_detector.detect(dst, estimate_tag_pose=True,
                                       camera_params=self.params,
                                       tag_size=self.tag_size)

        pose = None
        if len(tags) > 0:
            temp_mtx = detection2matrix(tags[0].pose_t, tags[0].pose_R)
            if str(tags[0].tag_id) in self.apriltags:
                # given tag pose in world coordinates and tag pose
                # in camera coordinates, perform transformation and return
                # pose of camera in world coordinates
                tag_mtx = self.apriltags[str(tags[0].tag_id)]
                temp_mtx = FLIP_MATRIX @ temp_mtx @ FLIP_MATRIX
                pose = tag_mtx @ np.linalg.inv(temp_mtx)
            else:
                print('Unknown Apriltag')

        out = {
            "pose": pose,
        }
        serializable_out = loads(dumps(
            out,
            cls=TransformedApriltagDetectorJSONEncoder
        ))

        # this publishes to subtopic (self.topic)
        self.publish({"data": serializable_out})


class TransformedApriltagDetectorJSONEncoder(JSONEncoder):
    """JSON Encoder helper for Apriltag Detector transformed attributes."""

    def default(self, o):
        """JSON Encoder helper for Apriltag Detector pose attributes."""
        if isinstance(o, (np.ndarray)):
            return list(o)

        return JSONEncoder.default(self, o)



"""
__init__.py: Service class for ARENA-robot.

Created by Perry Naseck on 1/25/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

import json
from datetime import datetime

from arena.device import Device
from paho.mqtt.client import MQTTMessage


class ArenaRobotService():
    """
    Service class for the ARENA.

    This class should be created/initialized as normal. Then check if
    async_service is True. If so, async_setup() should be run. If not, then
    setup() should be run. Lastly, run start() when the service should begin
    listening.
    """

    DEVICE_INSTANCE_TYPE = "unknown"

    # pylint: disable=too-many-arguments
    def __init__(self, instance_name: str, subtopic: str,
                 device_instance_type="unknown",
                 device_instance_prefix="",
                 async_service=False, interval_ms=-1):
        """Initialize the service class."""
        self.device_instance_type = device_instance_type
        instance_name = f"service_{device_instance_prefix}{instance_name}"
        self.instance_name = instance_name
        self.async_service_val = async_service
        self.interval_ms = interval_ms
        self.device = Device()
        self.topic = (f"{self.device.realm}/d/{self.device.namespace}/"
                      f"{self.device.device}/{subtopic}")
        self.device.message_callback_add(self.topic, self.msg_rx)

        print(f'Using topic {self.topic}')

    @property
    def async_service(self):
        """Assert if this is an async service."""
        return self.async_service_val

    @staticmethod
    def decode_payload(msg: MQTTMessage) -> dict:
        """Decode MQTTMessage as JSON."""
        try:
            payload_str = msg.payload.decode("utf-8", "ignore")
            payload = json.loads(json.loads(payload_str))
            return payload
        except json.JSONDecodeError as error:
            print("Malformed payload, ignoring:")
            print(error)
            return None

    def msg_rx(self, client, userdata, msg: MQTTMessage):
        """Receive messages."""

    def setup(self):
        """Set up sensor."""
        self.publish({"status": "initialized"})

    async def async_setup(self):
        """Set up sensor in async."""
        self.setup()

    def publish_msg_base(self):
        """Publish message base."""
        payload = {
            "device_instance_type": self.device_instance_type,
            "device_instance_name": self.instance_name,
            "timestamp": datetime.now().isoformat()[:-3]+"Z"
        }
        return payload

    def publish(self, msg):
        """Publish messages."""
        payload = self.publish_msg_base()
        payload.update({
            "msg": msg
        })
        payload = json.dumps(payload)
        self.device.publish(self.topic, payload)

    def fetch(self):
        """Fetch data."""
        self.publish({"data": "unknown"})

    async def async_fetch(self):
        """Fetch data in async."""
        self.fetch()

    def start(self):
        """Start fetching data."""
        self.publish({"status": "starting"})
        if self.async_service:
            self.device.run_async(self.async_fetch)
        elif self.interval_ms < 0:
            self.device.run_once(self.fetch)
        else:
            self.device.run_forever(self.fetch, interval_ms=self.interval_ms)
        self.device.run_tasks()



"""
licosa_serial.py: Definition for LiCosa IMU and lidar modules over serial.

Created by Perry Naseck on 2/22/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

import time
from dataclasses import asdict
from json import dumps, loads
from struct import unpack

from serial import Serial

from arenarobot.service.sensor import ArenaRobotServiceSensor
from licosa_py.packets import LiCosaJSONEncoder
from licosa_py.parser import LiCosaPacketParser


# pylint: disable=too-many-instance-attributes
class ArenaRobotServiceSensorLiCosaSerial(ArenaRobotServiceSensor):
    """LiCosa serial sensor class for the ARENA."""

    DEVICE_INSTANCE_SENSOR_TYPE = "licosa_serial"

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-statements
    def __init__(self, dev_path: str, baudrate: int = 230400, **kwargs):
        """Initialize the LiCosa serial sensor class."""
        self.dev_path = dev_path
        self.baudrate = baudrate

        self.serial_instance = None
        self.pkt = None

        sensor_type = (ArenaRobotServiceSensorLiCosaSerial
                       .DEVICE_INSTANCE_SENSOR_TYPE)
        super().__init__(
            device_instance_sensor_type=sensor_type,
            device_instance_prefix="licosa_serial_",
            **kwargs
        )

    def setup(self):
        """Set up LiCosa serial sensor."""
        self.serial_instance = Serial(port=self.dev_path,
                                      baudrate=self.baudrate,
                                      timeout=0)
        self.pkt = LiCosaPacketParser()
        time.sleep(1)
        super().setup()

    def fetch(self):
        """Fetch LiCosa serial data."""
        msg_bytes = self.serial_instance.read(size=32)

        # See thread here, this is the fastest way to iterate over bytes
        # https://stackoverflow.com/questions/14267452/iterate-over-individual-bytes-in-python-3
        for msg_byte in list(unpack(str(len(msg_bytes)) + 'c', msg_bytes)):
            pkt_out = self.pkt.next_byte(msg_byte)
            if pkt_out is not False:
                data = asdict(pkt_out)
                del data['data_buf']
                data = loads(dumps(data, cls=LiCosaJSONEncoder))
                self.publish({"data": data})



"""
__init__.py: Definitions for the occupancy grid.

Created by Perry Naseck on 6/28/21.

Copyright (c) 2021, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

from enum import Enum
from json import dumps, loads
from typing import NamedTuple

import numpy as np

from .const import (UPDATE_ACTION, UPDATE_ACTION_LEAVE, UPDATE_ACTION_OCCUPY,
                    UPDATE_LOC, UPDATE_OCCUPANT)


class Occupant:
    """Class to define an occupant on a grid."""

    def __init__(self, occupant_id):
        """Set up occupant."""
        self.occupant_id = occupant_id

    def __repr__(self):
        """Display occupant."""
        return f'Occupant(id={self.occupant_id})'

    def __eq__(self, other):
        """Equality of occupants."""
        return self.occupant_id == other.occupant_id

    def get_id(self):
        """Get the ID of the occupant."""
        return self.occupant_id


class Occupants(set):
    """Class to define a list of occupants."""


class OccupancyGridLocation(NamedTuple):
    """Class to define a location in the occupancy grid."""

    # pylint: disable=too-few-public-methods
    # Bug in pylint
    x: np.intp
    y: np.intp
    z: np.intp


class OccupancyGridAction(Enum):
    """Class to define acctions for an occupancy grid update."""

    OCCUPY = UPDATE_ACTION_OCCUPY
    LEAVE = UPDATE_ACTION_LEAVE


class OccupancyGridUpdate():
    """Class to define an occupancy grid update."""

    def __init__(self, update_str=None, update_data=None):
        """Set up occupancy grid update."""
        if update_str is None and update_data is None:
            raise ValueError("Either update_str or update_data must be set")
        if update_str is not None and update_data is not None:
            raise ValueError("Only update_str or update_data may be set"
                             "but not both")
        if update_data is not None:
            self.update_data = update_data
            self.update_str = self.encode_str(update_data)
        else:
            self.update_str = update_str
            self.update_data = self.decode_str(update_str)

    def __repr__(self):
        """Display occupancy grid update."""
        return f'OccupancyGridUpdate(action={self.update_data.action}, ' \
               f'loc={self.update_data.loc})'

    def __str__(self):
        """Display string representation of occupancy grid update."""
        return self.update_str

    def __eq__(self, other):
        """Equality of updates."""
        return self.update_data == other.update_data

    @staticmethod
    def validate(update_data: dict, raise_error=False) -> bool:
        """
        Validate an update dict.

        Note: This is a very simple validation, it only verifies
        the correct combnations of top level keys.
        TODO: Schema-based validation with voluptuous
        """
        if UPDATE_ACTION not in update_data:
            if raise_error:
                raise ValueError('All updates must have an action')
            return False
        if UPDATE_LOC not in update_data:
            if raise_error:
                raise ValueError('All updates must have a location')
            return False
        if (update_data[UPDATE_ACTION] == UPDATE_ACTION_LEAVE or
                update_data[UPDATE_ACTION] == UPDATE_ACTION_OCCUPY):
            if UPDATE_OCCUPANT not in update_data:
                if raise_error:
                    raise ValueError('This action requires an occupant')
                return False
        return True

    @staticmethod
    def encode_str(update_data: dict) -> str:
        """Encode an update as a string."""
        OccupancyGridUpdate.validate(update_data, raise_error=True)
        return dumps(update_data)

    @staticmethod
    def decode_str(update_str: str) -> dict:
        """Dencode an update to a dict."""
        data = loads(update_str)
        OccupancyGridUpdate.validate(data, raise_error=True)
        if UPDATE_LOC in data:
            loc = data[UPDATE_LOC]
            data[UPDATE_LOC] = OccupancyGridLocation(x=loc.x, y=loc.y, z=loc.z)
        if UPDATE_OCCUPANT in data:
            data[UPDATE_OCCUPANT] = Occupant(data[UPDATE_OCCUPANT])
        return data


class OccupancyGrid:
    """Class to define a grid of occupants."""

    def __init__(self, grid_id, size: np.intp):
        """Set up occupancy grid."""
        self.grid_id = grid_id
        self.size = size
        self.grid = np.full((size, size, size), Occupants)

    def __repr__(self):
        """Display occupancy grid."""
        return f'OccupancyGrid(id={self.grid_id}, size={self.size})'

    def num_occupying(self, loc: OccupancyGridLocation) -> int:
        """Get the number of occupants in a space in the grid."""
        return len(self.grid[loc.x][loc.y][loc.z])

    def occupy(self, loc: OccupancyGridLocation,
               occupant: Occupant) -> str:
        """
        Occupy a space in the grid with an occupant.

        loc (OccupancyGridLocation): location to occupy
        occupant (Occupant): occupant to place in coordinates
        ret_update (bool): return JSON update message
        """
        (self.grid[loc.x][loc.y][loc.z]).add(occupant)
        return OccupancyGridUpdate(update_data={
            UPDATE_ACTION: OccupancyGridAction.OCCUPY,
            UPDATE_LOC: loc,
            UPDATE_OCCUPANT: occupant
        })

    def leave(self, loc: OccupancyGridLocation,
              occupant: Occupant) -> OccupancyGridUpdate:
        """
        Leave a space in the grid with an occupant.

        loc (OccupancyGridLocation): location to leave
        occupant (Occupant): occupant to remove from coordinates
        ret_update (bool): return JSON update message
        """
        # discard() will not error in case gets out-of-sync or joins late
        (self.grid[loc.x][loc.y][loc.z]).discard(occupant)
        return OccupancyGridUpdate(update_data={
            UPDATE_ACTION: OccupancyGridAction.LEAVE,
            UPDATE_LOC: loc,
            UPDATE_OCCUPANT: occupant
        })

    def action(self, update: OccupancyGridUpdate) -> OccupancyGridUpdate:
        """Run an action given an update."""
        if update[UPDATE_ACTION] == OccupancyGridAction.OCCUPY:
            return self.occupy(loc=update[UPDATE_LOC],
                               occupant=update[UPDATE_OCCUPANT])
        if update[UPDATE_ACTION] == OccupancyGridAction.LEAVE:
            return self.leave(loc=update[UPDATE_LOC],
                              occupant=update[UPDATE_OCCUPANT])
        raise ValueError("Unknown action")



#!/usr/bin/env python
"""
licosa_py_basic.py: Get data from a LiCosa serial device.

Created by Perry Naseck on 2/22/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

import argparse
import asyncio
from dataclasses import asdict

# from aioserial import AioSerial
from serial import Serial

from licosa_py.parser import LiCosaPacketParser

async def main():
    parser = argparse.ArgumentParser(
        description='Get data from a LiCosa serial device.')
    parser.add_argument('dev_path', metavar='/dev/*', type=str,
                        help='serial device path')
    parser.add_argument('baudrate', nargs='?', type=int, default=230400,
                        help='serial device baudrate')
    args = parser.parse_args()

    # aioserial_instance = AioSerial(port=args.dev_path, baudrate=args.baudrate)
    serial_instance = Serial(port=args.dev_path, baudrate=args.baudrate)
    pkt = LiCosaPacketParser()

    while True:
        # msg_byte = await aioserial_instance.read_async(size=1)
        msg_byte = serial_instance.read(size=1)

        pkt_out = pkt.next_byte(msg_byte)
        if pkt_out is not False:
            data = asdict(pkt_out)
            print(data)

if __name__=="__main__":
    res = asyncio.run(main())
    exit(res)



#!/usr/bin/env python3
"""
arena_video_call_chrome.py: Connect to an ARENA video call in Chrome.

Created by Perry Naseck on 7/1/21.

Copyright (c) 2021, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

# NOTE: This example requires the Selenium Chrome Driver
#       On Raspberry Pi you can get it with:
#         sudo apt-get install chromium-chromedriver
#       Works well on Rapsbian Lite.

import sys
from platform import node

from arena import Scene
from selenium.webdriver import Chrome, ChromeOptions

from arenavideocall import VideoCall


def main() -> int:
    """Start a video call example."""
    if len(sys.argv) > 1 and sys.argv[1] == '--help':
        print(f'Usage: {__file__} [object_id]\n'
              'Specifcy arena-py environment vairables to select the '
              'scene\n'
              'See https://arena.conix.io'
              '/content/python/#running-from-the-command-line')
        return 0

    object_id = None
    if len(sys.argv) > 1:
        object_id = sys.argv[1]

    options = ChromeOptions()
    options.headless = True
    options.add_argument("--use-fake-ui-for-media-stream=1")
    options.add_argument("--enable-accelerated-video-decode")
    options.add_argument("--enable-gpu-rasterization")
    options.add_argument("--ignore-gpu-blocklist")

    scene = Scene(video=True)

    call = VideoCall(scene, Chrome, options)

    call.open()
    call.change_display_name(node())  # Hostname
    call.set_audio_mute(True)
    call.set_video_mute(False)

    call.wait_for_join()  # Wait for call to finish joining

    print('Jitsi User ID:', call.get_user_id())
    print('Jitsi Display Name:', call.get_display_name())
    print('Jitsi Room Name:', call.get_room_name())
    print('Jitsi Audio Muted:', call.get_audio_mute())
    print('Jitsi Video Muted:', call.get_video_mute())
    print('Jitsi Video Track Label:', call.get_video_track_label())
    print('Jitsi Video Resolution:', call.get_video_resolution())

    @scene.run_once
    def update_video_object():
        if object_id is not None:
            print(f"Updating object '{object_id}' to use new video")
            obj = scene.get_persisted_obj(object_id)
            call.update_object_video(obj)

    scene.run_tasks()

    return 0


if __name__ == "__main__":
    RES = main()
    sys.exit(RES)



"""
beluga_serial.py: Definition for Beluga UWB modules over serial.

Created by Perry Naseck on 1/27/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

import time
from typing import Optional

from aioserial import AioSerial

from arenarobot.service.sensor import ArenaRobotServiceSensor


# pylint: disable=too-many-instance-attributes
class ArenaRobotServiceSensorBelugaSerial(ArenaRobotServiceSensor):
    """Beluga serial sensor class for the ARENA."""

    DEVICE_INSTANCE_SENSOR_TYPE = "beluga_serial"

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-statements
    def __init__(self, dev_path: str, beluga_id: int, baudrate: int = 115200,
                 beluga_rate: int = 10, beluga_channel: Optional[int] = None,
                 beluga_tx_power: Optional[int] = None,
                 beluga_timeout: Optional[int] = None,
                 beluga_twr_mode: Optional[int] = None,
                 beluga_led_mode: Optional[int] = None,
                 **kwargs):
        """Initialize the Beluga serial sensor class."""
        self.dev_path = dev_path
        self.baudrate = baudrate
        self.beluga_id = beluga_id
        self.beluga_rate = beluga_rate
        self.beluga_channel = beluga_channel
        self.beluga_tx_power = beluga_tx_power
        self.beluga_timeout = beluga_timeout
        self.beluga_twr_mode = beluga_twr_mode
        self.beluga_led_mode = beluga_led_mode

        self.aioserial_instance = None

        sensor_type = (ArenaRobotServiceSensorBelugaSerial
                       .DEVICE_INSTANCE_SENSOR_TYPE)
        super().__init__(
            device_instance_sensor_type=sensor_type,
            device_instance_prefix="beluga_serial_",
            async_service=True, **kwargs
        )

    async def async_setup(self):
        """Set up Beluga serial sensor."""
        self.aioserial_instance = AioSerial(port=self.dev_path,
                                            baudrate=self.baudrate)

        def msg_in_clear():
            # Wait and clear all of the messages that come in
            # Python seems bad at clearing the buffers the first time
            for _ in range(2):
                while self.aioserial_instance.in_waiting > 0:
                    self.aioserial_instance.reset_input_buffer()
                time.sleep(1)

        async def wait_for_msg(wait_msg: str):
            while True:
                msg_bytes = await self.aioserial_instance.readline_async()
                msg = msg_bytes.decode(errors="ignore")
                if msg.startswith(wait_msg):
                    break

        msg_in_clear()

        set_id_msg = "AT+RESET\n".encode()
        await self.aioserial_instance.write_async(set_id_msg)
        await wait_for_msg("Reset OK")
        time.sleep(1)

        set_id_msg = "AT+REBOOT\n".encode()
        await self.aioserial_instance.write_async(set_id_msg)

        msg_in_clear()

        set_id_msg = f"AT+ID {self.beluga_id}\n".encode()
        await self.aioserial_instance.write_async(set_id_msg)
        await wait_for_msg("OK")

        set_boot_mode_msg = "AT+BOOTMODE 0\n".encode()
        await self.aioserial_instance.write_async(set_boot_mode_msg)
        await wait_for_msg("Bootmode: 0 OK")

        if self.beluga_rate is not None:
            set_rate_msg = f"AT+RATE {self.beluga_rate}\n".encode()
            await self.aioserial_instance.write_async(set_rate_msg)
            await wait_for_msg(f"Rate: {self.beluga_rate} OK")

        if self.beluga_channel is not None:
            set_channel_msg = f"AT+CHANNEL {self.beluga_channel}\n".encode()
            await self.aioserial_instance.write_async(set_channel_msg)
            await wait_for_msg("OK")

        if self.beluga_tx_power is not None:
            set_tx_power_msg = f"AT+TXPOWER {self.beluga_tx_power}\n".encode()
            await self.aioserial_instance.write_async(set_tx_power_msg)
            await wait_for_msg("OK")

        if self.beluga_timeout is not None:
            set_timeout_msg = f"AT+TIMEOUT {self.beluga_timeout}\n".encode()
            await self.aioserial_instance.write_async(set_timeout_msg)
            await wait_for_msg("OK")

        set_stream_mode_msg = "AT+STREAMMODE 0\n".encode()
        await self.aioserial_instance.write_async(set_stream_mode_msg)
        await wait_for_msg("OK")

        if self.beluga_twr_mode is not None:
            set_twr_mode_msg = f"AT+TWRMODE {self.beluga_twr_mode}\n".encode()
            await self.aioserial_instance.write_async(set_twr_mode_msg)
            await wait_for_msg("OK")

        if self.beluga_led_mode is not None:
            set_led_mode_msg = f"AT+LEDMODE {self.beluga_led_mode}\n".encode()
            await self.aioserial_instance.write_async(set_led_mode_msg)
            await wait_for_msg("OK")

        start_ble_msg = "AT+STARTBLE\n".encode()
        await self.aioserial_instance.write_async(start_ble_msg)
        await wait_for_msg("OK")
        time.sleep(0.25)

        start_uwb_msg = "AT+STARTUWB\n".encode()
        await self.aioserial_instance.write_async(start_uwb_msg)
        await wait_for_msg("OK")

        time.sleep(1)
        await super().async_setup()

    async def async_fetch(self):
        """Fetch Beluga serial data."""
        self.aioserial_instance.reset_input_buffer()
        while True:
            msg_bytes = await self.aioserial_instance.readline_async()
            msg = msg_bytes.decode(errors="ignore")
            if msg[0] == "#":
                # Skip publishing comment messages
                continue
            msg_split = msg.split(", ")
            if len(msg_split) < 4:
                # Skip garbled messages
                continue
            rx_id, rx_range, rx_rssi, rx_timestamp = msg_split
            data = {
                "rx_nodes": [{
                    "src": int(self.beluga_id),
                    "dst": int(rx_id),
                    "distance": float(rx_range),
                    "ble_rssi": int(rx_rssi),
                    "counter": int(rx_timestamp)
                }]
            }
            self.publish({"data": data})



"""
vl53l5cx.py: Definition for VL53L5CX sensors.

Created by Perry Naseck on 1/21/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

import time
from json import dumps, loads
from typing import Sequence

import numpy as np
from periphery import GPIO

from arenarobot.service.sensor import ArenaRobotServiceSensor
from vl53l5cx_py.helpers import VL53L5CXJSONEncoder

# Optional imports allow for constants from this module to be referenced
# and for the service to be recognized without all of the required
# dependencies installed need for starting this service.
imports_missing = []
try:
    from vl53l5cx_py.driver import VL53L5CX
except ImportError as err:
    VL53L5CX = err
    imports_missing.append(err)


class ArenaRobotServiceSensorVL53L5CX(ArenaRobotServiceSensor):
    """VL53L5CX sensor class for the ARENA."""

    DEVICE_INSTANCE_SENSOR_TYPE = "vl53l5cx"

    def __init__(self, dev_path: str, gpio_path: str, rst_pin: int,
                 lpn_pins: Sequence[int], **kwargs):
        """Initialize the VL53L5CX sensor class."""
        # Check for missing imports
        if len(imports_missing) > 0:
            raise Exception(imports_missing)

        self.dev_path = dev_path
        self.gpio_path = gpio_path
        self.rst_pin = rst_pin
        self.lpn_pins = lpn_pins
        self.sensors_setup = [None] * len(self.lpn_pins)
        self.sensors = [None] * len(self.lpn_pins)

        sensor_type = (ArenaRobotServiceSensorVL53L5CX
                       .DEVICE_INSTANCE_SENSOR_TYPE)
        super().__init__(
            device_instance_sensor_type=sensor_type,
            device_instance_prefix="vl53l5cx_",
            **kwargs
        )

    def setup(self):
        """Set up VL53L5CX sensor."""
        rst_pin = GPIO(self.gpio_path, self.rst_pin, "out")
        rst_pin.write(False)

        for i, pin_num in enumerate(self.lpn_pins):
            pin = GPIO(self.gpio_path, pin_num, "out")
            pin.write(False)
            addr = VL53L5CX.DEFAULT_ADDR + np.uint16(2 * i)
            self.sensors_setup[i] = ((i, pin, addr,))
            print(f'Will setup sensor {i}, addr {addr:02x} on pin {pin_num}')

        rst_pin.write(True)
        time.sleep(1)
        rst_pin.write(False)
        time.sleep(1)

        for sensor_setup in self.sensors_setup:
            i, pin, addr = sensor_setup
            pin.write(True)
            time.sleep(1)
            self.sensors[i] = VL53L5CX(self.dev_path, target_addr=addr)
            pin.write(False)
            time.sleep(1)

        for sensor_setup in self.sensors_setup:
            i, pin, addr = sensor_setup
            pin.write(True)
            time.sleep(1)
            self.sensors[i].start_ranging()

        super().setup()

    def fetch(self):
        """Fetch VL53L5CX data."""
        data = []
        for sensor in self.sensors:
            ranges = sensor.get_range()
            data.append(ranges)
        data = loads(dumps(data, cls=VL53L5CXJSONEncoder))
        self.publish({"data": data})



#!/usr/bin/env python
"""
setup.py: Package setup file.

Created by Perry Naseck on 6/28/21.

Copyright (c) 2021, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

from glob import glob
from Cython.Build import cythonize
import pathlib
from setuptools import setup, Extension, find_packages
import numpy
from sys import platform

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

extensions = []

# vl53l5cx_py only available on Linux
if platform.startswith('linux'):
    extensions.append(Extension("vl53l5cx_py.driver",
              (["./vl53l5cx_py/cython/vl53l5cx_py_wrapper.pyx"] +
               glob("./vl53l5cx_py/src/*.c") +
               glob("./vl53l5cx_py/VL53L5CX_Linux_driver_1.1.2/user/uld-driver/src/*.c") +
               glob("./vl53l5cx_py/VL53L5CX_Linux_driver_1.1.2/user/platform/*.c")),
              include_dirs=["./vl53l5cx_py/include",
                            "./vl53l5cx_py/VL53L5CX_Linux_driver_1.1.2/user/uld-driver/inc",
                            "./vl53l5cx_py/VL53L5CX_Linux_driver_1.1.2/user/platform",
                            numpy.get_include()]
    ))
else:
    extensions.append(Extension("vl53l5cx_py.driver",
              (["./vl53l5cx_py/cython/vl53l5cx_py_wrapper_disable.pyx"]),
              include_dirs=[]
    ))

setup(
    name="arena-robot",
    description="Robot integration runtime for the ARENA",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='BSD-3-Clause',
    author="Conix Research Center",
    author_email="info@conix.io",
    python_requires='>=3.8, <4',
    packages=find_packages(),
    ext_modules=cythonize(extensions, language_level="3"),
    zip_safe=False,
    platforms=["any"],
    url="https://github.com/arenaxr/arena-robot",
    project_urls={
        'Bug Reports': 'https://github.com/arenaxr/arena-robot/issues',
        'Source': 'https://github.com/arenaxr/arena-robot',
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Robot Framework',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: C',
        'Programming Language :: Cython',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Hardware'
    ],
    keywords='arena robot ar vr robotics drone lidar vl53l5cx stm selenium jitsi video videocall conference',
    entry_points = {
        'console_scripts': [
            'arena-video-call-chrome=arenavideocall.scripts.arena_video_call_chrome:main',
            'arena-robot-service=arenarobot.scripts.arena_robot_service:main'
        ],
    },
    install_requires=[
        'aioserial>=1,<2',
        'arena-py>=0.1.43,<1',
        'scipy>=1.8',
        'numpy>=1,<2',
        'pyserial>=3,<4',
        'python-periphery>=2,<3',
        'selenium>=3,<4',
        'setuptools_scm>=6.0',
        'slam3d>=0,<1',
        'transformations==2021.6.6'
    ],
    extras_require={
        'dev': [
            'bandit>=1,<2',
            'flake8>=3,<4',
            'flake8-bugbear>=21,<22',
            'isort>=5,<6',
            'pydocstyle>=6,<7',
            'pylint>=2,<3'
        ],
        'realsense': [
            'pyrealsense2>=2,<3'
        ],
        'apriltag_detector': [
            'dt-apriltags>=3,<4',
            # issue with pylint later
            # see https://github.com/opencv/opencv-python/issues/570#issuecomment-1151984301
            'opencv-python==4.5.5.64'
        ]
    }
)



#####################################################
##          librealsense T265 streams test         ##
#####################################################
# This assumes .so file is found on the same directory
import pyrealsense2.pyrealsense2 as rs

# Prettier prints for reverse-engineering
from pprint import pprint
import numpy as np

# Get realsense pipeline handle
pipe = rs.pipeline()

# Print all connected devices and find the T265
devices = rs.context().devices
for i in range(len(devices)):
    print('Found device:', devices[i].get_info(rs.camera_info.name), ', with serial number: ', devices[i].get_info(rs.camera_info.serial_number))

# Configure the pipeline
cfg = rs.config()

# Prints a list of available streams, not all are supported by each device
print('Available streams:')
pprint(dir(rs.stream))

# Enable streams you are interested in
cfg.enable_stream(rs.stream.pose) # Positional data (translation, rotation, velocity etc)
cfg.enable_stream(rs.stream.fisheye, 1) # Left camera
cfg.enable_stream(rs.stream.fisheye, 2) # Right camera

# Start the configured pipeline
pipe.start(cfg)

try:
    while(1):
        frames = pipe.wait_for_frames()

        # Left fisheye camera frame
        left = frames.get_fisheye_frame(1)
        left_data = np.asanyarray(left.get_data())

        # Right fisheye camera frame
        right = frames.get_fisheye_frame(2)
        right_data = np.asanyarray(right.get_data())

        print('Left frame', left_data.shape)
        print('Right frame', right_data.shape)

        # Positional data frame
        pose = frames.get_pose_frame()
        if pose:
            pose_data = pose.get_pose_data()
            print("\nFrame number: %5.0f" % (pose.frame_number))
            print("Position xyz: % 2.4f % 2.4f % 2.4f" % (pose_data.translation.x, pose_data.translation.y, pose_data.translation.z))
            print("Velocity xyz: % 2.4f % 2.4f % 2.4f" % (pose_data.velocity.x, pose_data.velocity.y, pose_data.velocity.z))
            print("Accelera xyz: % 2.4f % 2.4f % 2.4f" % (pose_data.acceleration.x, pose_data.acceleration.y, pose_data.acceleration.z))
            print("Quatern xyzw: % 2.4f % 2.4f % 2.4f % 2.4f" % (pose_data.rotation.x, pose_data.rotation.y, pose_data.rotation.z, pose_data.rotation.w))
finally:
    pipe.stop()



"""
t265.py: Transformations for T265 input.

Created by Perry Naseck on 6/1/22.

Copyright (c) 2022, The CONIX Research Center
All rights reserved.

This source code is licensed under the BSD-3-Clause license found in the
LICENSE file in the root directory of this source tree.
"""

from json import JSONEncoder, dumps, loads

import numpy as np
from paho.mqtt.client import MQTTMessage
from transformations import euler_matrix, quaternion_matrix

from arenarobot.service.processor import ArenaRobotServiceProcessor
from arenarobot.service.sensor.t265 import ArenaRobotServiceSensorT265


# pylint: disable=too-many-instance-attributes
class ArenaRobotServiceProcessorT265(ArenaRobotServiceProcessor):
    """T265 processor class for the ARENA."""

    DEVICE_INSTANCE_PROCESSOR_TYPE = "t265"

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-statements
    def __init__(self, sensor_t265_topic: str,
                 sensor_t265_instance_name: str = None,
                 camera_orientation: int = 0,
                 scale_factor: float = 1.0,
                 jump_threshold_translation: float = 0.1,
                 jump_threshold_velocity: float = 20.0,
                 **kwargs):
        """Initialize the Y265 processor class."""
        self.sensor_t265_topic = sensor_t265_topic
        self.sensor_t265_instance_name = sensor_t265_instance_name
        self.camera_orientation = camera_orientation
        self.scale_factor = float(scale_factor)
        self.jump_threshold_translation = float(jump_threshold_translation)
        self.jump_threshold_velocity = float(jump_threshold_velocity)

        if (not isinstance(self.camera_orientation, int) or
                not 0 <= self.camera_orientation <= 3):
            raise ValueError('Unknown camera orientation!')

        self.fetched_once = False
        self.prev_data = None
        self.just_initialized = False
        self.h_aeroref_t265ref = None
        self.h_t265body_aerobody = None

        processor_type = (ArenaRobotServiceProcessorT265
                          .DEVICE_INSTANCE_PROCESSOR_TYPE)
        super().__init__(
            device_instance_processor_type=processor_type,
            device_instance_prefix="t265_",
            **kwargs
        )

    def setup(self):
        """Set up T265 processor."""
        # Forward, USB port to the right
        if self.camera_orientation == 0:
            self.h_aeroref_t265ref = np.array(
                [[0, 0, -1, 0], [1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 0, 1]])
            self.h_t265body_aerobody = np.linalg.inv(self.h_aeroref_t265ref)
        # Downfacing, USB port to the right
        elif self.camera_orientation == 1:
            self.h_aeroref_t265ref = np.array(
                [[0, 0, -1, 0], [1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 0, 1]])
            self.h_t265body_aerobody = np.array(
                [[0, 1, 0, 0], [1, 0, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
        # 45degree forward
        elif self.camera_orientation == 2:
            self.h_aeroref_t265ref = np.array(
                [[0, 0, -1, 0], [1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 0, 1]])
            self.h_t265body_aerobody = (
                euler_matrix(np.pi/4, 0, 0)).dot(
                    np.linalg.inv(self.h_aeroref_t265ref))
        # Upward facing, USB port to the right
        elif self.camera_orientation == 3:
            self.h_aeroref_t265ref = np.array(
                [[0, 0, -1, 0], [1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 0, 1]])
            self.h_t265body_aerobody = np.array(
                [[0, 1, 0, 0], [-1, 0, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
        super().setup()

    def fetch(self):
        """Fetch T265 data."""
        # Ensure not run more than once (sets up callbacks)
        if self.fetched_once:
            print("T265 processor should have interval of -1!")
            return
        self.fetched_once = True

        # pylint: disable=unused-argument
        # pylint: disable=too-many-locals
        def process_t265(client, userdata, msg: MQTTMessage):
            """Process a T265 sensor message."""
            payload = self.decode_payload(msg)
            if payload is None:
                return

            # Ignore messages from other sensor types
            if (payload["device_instance_sensor_type"] !=
                    ArenaRobotServiceSensorT265.DEVICE_INSTANCE_SENSOR_TYPE):
                return

            # If set, ignore other instances of T265 sensors
            if (self.sensor_t265_instance_name is not None and
                    payload["device_instance_name"] !=
                    self.sensor_t265_instance_name):
                return

            if ("status" in payload["msg"] and
                    payload["msg"]["status"] == "initialized"):
                print("Detected T265 sensor initialization")
                self.just_initialized = True

            if "data" not in payload["msg"]:
                print("Skipping missing data for",
                      payload["device_instance_name"],
                      payload["msg"])
                return
            data = payload["msg"]["data"]

            # Skip errors
            if "error" in data:
                print("Skipping errored data for",
                      payload["device_instance_name"])
                return
            if "pose" not in data:
                print("Pose missing from data for",
                      payload["device_instance_name"])
                return

            pose = data["pose"]
            rotation = pose["rotation"]
            translation = pose["translation"]
            velocity = pose["velocity"]

            # Source:
            # https://discuss.ardupilot.org/t/integration-of-ardupilot-and-vio-tracking-camera-part-4-non-ros-bridge-to-mavlink-in-python/44001/45
            # In transformations, Quaternions w+ix+jy+kz are represented as
            # [w, x, y, z]!
            h_t265_t265body = quaternion_matrix([
                rotation["w"],
                rotation["x"],
                rotation["y"],
                rotation["z"]])
            h_t265_t265body[0][3] = translation["x"] * self.scale_factor
            h_t265_t265body[1][3] = translation["y"] * self.scale_factor
            h_t265_t265body[2][3] = translation["z"] * self.scale_factor

            # Transform to aeronautic coordinates (body AND reference frame!)
            h_aeroref_aerobody = self.h_aeroref_t265ref.dot(
                h_t265_t265body.dot(self.h_t265body_aerobody))

            # Calculate GLOBAL XYZ speed (speed from T265 is already GLOBAL)
            v_aeroref_aerobody = quaternion_matrix([1, 0, 0, 0])
            v_aeroref_aerobody[0][3] = velocity["x"]
            v_aeroref_aerobody[1][3] = velocity["y"]
            v_aeroref_aerobody[2][3] = velocity["z"]
            v_aeroref_aerobody = self.h_aeroref_t265ref.dot(v_aeroref_aerobody)

            # Check for pose jump
            if self.prev_data is not None:
                prev_translation = self.prev_data["pose"]["translation"]
                prev_velocity = self.prev_data["pose"]["velocity"]
                delta_translation = [
                    translation["x"] - prev_translation["x"],
                    translation["y"] - prev_translation["y"],
                    translation["z"] - prev_translation["z"]
                ]
                delta_velocity = [
                    velocity["x"] - prev_velocity["x"],
                    velocity["y"] - prev_velocity["y"],
                    velocity["z"] - prev_velocity["z"]
                ]
                delta_translation_norm = np.linalg.norm(delta_translation)
                delta_velocity_norm = np.linalg.norm(delta_velocity)

                # Pose jump is indicated when position changes abruptly. The
                # behavior is not well documented yet (as of
                # librealsense 2.34.0)
                jump_detected_translation = bool(
                    delta_translation_norm > self.jump_threshold_translation
                )
                jump_detected_velocity = bool(
                    delta_velocity_norm > self.jump_threshold_velocity
                )
            else:
                jump_detected_translation = False
                jump_detected_velocity = False
                delta_translation_norm = 0.0
                delta_velocity_norm = 0.0

            out = {
                "h_t265_t265body": h_t265_t265body,
                "h_aeroref_aerobody": h_aeroref_aerobody,
                "v_aeroref_aerobody": v_aeroref_aerobody,
                "jump_translation_detected": jump_detected_translation,
                "jump_velocity_detected": jump_detected_velocity,
                "delta_translation_norm": delta_translation_norm,
                "delta_velocity_norm": delta_velocity_norm,
                "just_initialized": self.just_initialized
            }
            self.prev_data = {
                "pose": pose,
                "out": out
            }
            serializable_out = loads(dumps(
                out,
                cls=TransformedT265PoseJSONEncoder
            ))
            self.publish({"data": serializable_out})
            self.just_initialized = False

        self.device.message_callback_add(self.sensor_t265_topic, process_t265)


class TransformedT265PoseJSONEncoder(JSONEncoder):
    """JSON Encoder helper for T265 transformed attributes."""

    def default(self, o):
        """JSON Encoder helper function for T265 pose attributes."""
        if isinstance(o, (np.ndarray)):
            return list(o)

        return JSONEncoder.default(self, o)
